#include <Trade/Trade.mqh>
#include <Trade/PositionInfo.mqh>
#include <Trade/AccountInfo.mqh>
#include <Trade/DealInfo.mqh>
#include <News.mqh>
#include <Trade/SymbolInfo.mqh>
#include <Telegram.mqh>
#include <Comment.mqh>
#include <Telegram.mqh>
#include <StringUtils.mqh>


enum ordertype

{
    BUY=1,

    SELL=2,

    BUY_LIMIT=3,

    SELL_LIMIT=4,

    BUY_STOP=5,

    SELL_STOP=6,

    BUY_STOP_LIMIT=7,

    SELL_STOP_LIMIT=8,

    BUY_LIMIT_STOP=9,

    SELL_LIMIT_STOP=10,
    
    NONE=11

};

enum curr_enum

{

    Base, //Base Currency.

    Profit, // Quote Currency.

    All // Both.

};

enum special_news_enum

{

    High_Importance, // High Importance news.

    Medium_Importance, //Medium Importance news.

    Low_Importance, // Low Importance news.

    Specific_News // Specified News Events.

};

enum hedge_type
{
    allow_hedge=0, //Allow to have both buy and sell active positions.
    close_opposite=1, //Close current position on opening of opposite position.
    await_close=2 //Don't open positions in opposite direction if there's an active position.
};

struct NewsData

  {

   datetime          ReleaseTime; // Release time.

   string            NewsName; // Title of News.

   string            NewsCurrency; // Currency Affected.

   ENUM_CALENDAR_EVENT_IMPORTANCE        Importance; //Importance of news.

  };
  
enum close_positions
{
    close_all=1, //Close all open positions.
    close_profit=2,//Close all profitable positions.
    close_loss=3,//Close all losing positions.
    leave_be=4, //Close all except positions on breakeven.
    leave_all=5//Leave all positions.
};

struct order
{
    double price;
    double limit_price;
    double take_profit;
    double stoploss;
    ordertype type;
    datetime expiration;
    datetime setup_time;
    string comment;
    string symbol;
    double breakeven;
    int index;
    bool active;
    long signal_id;
    double tps[];
};
 
enum enum_tps
{
    TP1=0, //TakeProfit 1
    TP2=1,//TakeProfit 2
    MAX_TP //Highest Takeprofit Level in signal.
};

enum signal_action
{
    open_new=0,
    be=1,
    close=2,
    modifytp=3,
    close_half=4,
    be2=5
};

struct signal
{
    string symbol;
    string _comment;
    ordertype type;
    double sl;
    double entry;
    double tp;
    long signal_id;
    long tickets[];
    long message_id;
    long message_ids[];
    datetime signal_time;
    double tps[];
};


const ENUM_TIMEFRAMES _periods[] = {PERIOD_M1,PERIOD_M5,PERIOD_M15,PERIOD_M30,PERIOD_H1,PERIOD_H4,PERIOD_D1,PERIOD_W1,PERIOD_MN1};
//+------------------------------------------------------------------+
//|   CMyBot                                                         |
//+------------------------------------------------------------------+
class CMyBot: public CCustomBot
{
private:
   string            m_symbol;
   ENUM_TIMEFRAMES   m_period;

public:
   //+------------------------------------------------------------------+
   string user_chat;


       
   bool ProcessAction(signal &ref_signal, signal_action order=0, string text="")
   {
       CTrade trade;
       CPositionInfo info;
       string return_text;
       
       
       string symbol=Instrument(FindW(text,true),text);
       double sl=0,tp=0,entry=0,cmp=0,tps[];
       GetStops(text,cmp,sl,tp,tps,entry);
       //ArrayPrint(signals);
       
       
       if((order==0)) //Create New Order(s).
       {
           ref_signal.sl=sl;
           if(ref_signal.symbol=="") ref_signal.symbol=symbol;
           ref_signal.tp=tp;
           ArrayResize(ref_signal.tps,ArraySize(tps));
           for(int i=0; i<ArraySize(tps); i++) ref_signal.tps[i]=tps[i];
           //ref_signal.tps=tps;
           ref_signal.entry=entry;
           ref_signal.type=NONE;
           Print("Entry: ",ref_signal.entry);
           ref_signal.type=GetType(text,ref_signal,return_text);
           //--
           if((ref_signal.type==NONE) || (ref_signal.symbol==""))
           {
               return false;
           }
           else if(ref_signal.sl==0 || ref_signal.tp==0)
           {
               bool invalid=false;
               string blacklist_words="running hit fly moon rocket waterfall done move summary invalid valid";
               string blacklists[];
               StringSplit(blacklist_words," ",blacklists);
               int i=ArraySize(blacklists);
               while(--i>=0) invalid = StringContains(text,blacklists[i])? true:invalid;
               if(invalid) return false;
           }
           int i=ArraySize(signals);
           ArrayResize(signals, ArraySize(signals)+1);
           signals[i]=ref_signal;
           //signals[i].signal_id=signal_id;
           CreateOrder(signals[i]);
           bot.SendMessage(bot.user_chat, return_text);
           return true;
       }
       
       int n=ArraySize(signals[SignalIndex(ref_signal.signal_id)].tickets);
       ArrayPrint(signals[SignalIndex(ref_signal.signal_id)].tickets);
       //
       if(order==2) //Close all.
       {
           
           for(int i=0; i<n; i++)
           {
               long ticket=signals[SignalIndex(ref_signal.signal_id)].tickets[i];
               info.SelectByTicket(ticket);
               string dir=info.PositionType()==POSITION_TYPE_BUY? " BUY":" SELL";
               return_text=info.Symbol() + dir + " Closed!";
               //--
               if(trade.PositionClose(ticket)) bot.SendMessage(bot.user_chat,return_text);
               bot.SendMessage(bot.user_chat,Trade.ResultRetcodeDescription());
           }
       } 
       else if(order==3)//Modify takeprofit.
       {
           
           ref_signal.tp=tp;
           if(tp==0) tp=GetDoubleValue(text);
           if(tp==0) return false;
           for(int i=0; i<n; i++)
           {
               long ticket=signals[SignalIndex(ref_signal.signal_id)].tickets[i];
               info.SelectByTicket(ticket);
               sl=info.StopLoss();
               string dir=info.PositionType()==POSITION_TYPE_BUY? " BUY":" SELL";
               return_text=info.Symbol() + dir + " Closed!";
               //--
               if(trade.PositionModify(ticket,sl,tp)) bot.SendMessage(bot.user_chat,return_text);
               bot.SendMessage(bot.user_chat,Trade.ResultRetcodeDescription());
           }
       }
       else if(order==4)//Close half.
       {
           
           for(int i=0; i<n; i++)
           {
               long ticket=signals[SignalIndex(ref_signal.signal_id)].tickets[i];
               info.SelectByTicket(ticket);
               double volume=info.Volume();
               string dir=info.PositionType()==POSITION_TYPE_BUY? " BUY":" SELL";
               return_text="Partials profits taken on " + info.Symbol() +dir;
               //--
               if(trade.PositionClosePartial(ticket,volume)) {bot.SendMessage(bot.user_chat,return_text); signals[SignalIndex(ref_signal.signal_id)].tickets[n]=trade.RequestPosition();}
               bot.SendMessage(bot.user_chat,Trade.ResultRetcodeDescription());
           }
       }
       else if(order==1)//Modify stoploss.
       {
           
           ref_signal.sl=sl;
           if(sl==0) sl=GetDoubleValue(text);
           for(int i=0; i<n; i++)
           {
               long ticket=signals[SignalIndex(ref_signal.signal_id)].tickets[i];
               info.SelectByTicket(ticket);
               if(sl==0) sl=info.StopLoss();
               tp=info.TakeProfit();
               string dir=info.PositionType()==POSITION_TYPE_BUY? " BUY":" SELL";
               return_text="StopLoss modified for" + info.Symbol() + dir;
               //--
               if(trade.PositionModify(ticket,sl,tp)) bot.SendMessage(bot.user_chat,return_text);
               bot.SendMessage(bot.user_chat,Trade.ResultRetcodeDescription());
           }
       }
       else if(order==5)//Close half; Set Breakeven
       {
           
           for(int i=0; i<n; i++)
           {
               long ticket=signals[SignalIndex(ref_signal.signal_id)].tickets[i];
               info.SelectByTicket(ticket);
               double volume=info.Volume();
               tp=info.TakeProfit();
               sl=info.PriceOpen();
               string dir=info.PositionType()==POSITION_TYPE_BUY? " BUY":" SELL";
               return_text="Breakeven set for" + info.Symbol() + dir;
               //--
               if(trade.PositionModify(ticket,sl,tp)) bot.SendMessage(bot.user_chat,return_text);
               return_text="Partials taken on" + info.Symbol() + dir;
               if(trade.PositionClosePartial(ticket,volume)) {bot.SendMessage(bot.user_chat,return_text); signals[SignalIndex(ref_signal.signal_id)].tickets[n]=trade.RequestPosition();}
               bot.SendMessage(bot.user_chat,Trade.ResultRetcodeDescription());
           }
       }
       return true;
   }
   

   ordertype GetType(string &msg, signal &new_signal, string &comment)
   {
       string symbol = new_signal.symbol;
       double cmp = SymbolInfoDouble(symbol, SYMBOL_BID);
       long spread = SymbolInfoInteger(symbol, SYMBOL_SPREAD);
       long freeze = SymbolInfoInteger(symbol, SYMBOL_TRADE_FREEZE_LEVEL);
       double point = SymbolInfoDouble(symbol, SYMBOL_POINT);
       double freeze_point = (freeze + spread) * point;
       double sl=new_signal.sl, tp=new_signal.tp, entry=new_signal.entry;
       bool valid = no_sl && sl<=0;
       
       int code = FindW(msg, true);
       
       //A quick fix
       if(code==0)
       {
           if((entry!=0) && (sl!=0))
           {
               if((entry>cmp) && (entry>sl)) code = 20;
               if((entry<cmp) && (entry<sl)) code = 23;
               if((entry<cmp) && (entry>sl)) code = 19;
               if((entry>cmp) && (entry<sl)) code = 22;
           }
           else if(sl!=0)
           {
               if(cmp<sl) code = 21;
               if(cmp>sl) code = 18;
           }
       }
       
       bool instant = (code==18) || (code==21);
       if(!instant)
       {
           if(fabs(entry-cmp) > freeze_point)
           {
               double high =iHigh(symbol, PERIOD_M6, 0);
               double low = iLow(symbol, PERIOD_M6, 0);
               Sleep(1000);
               double vol=micro?10:0.1;
               if((code==19) && (CostPips(symbol, entry, cmp, vol)<1.0) && (cmp>low) && (low<=entry)) {code=18; instant=true;}
               if((code==22) && (CostPips(symbol, entry, cmp, vol)<1.0) && (cmp<high) && (high>=entry)) {code=21; instant=true;}
           }
           if(!instant) valid = valid && entry!=0;
       }
       
       ordertype signtype = NONE;
       string p=instant? string(cmp) : string(entry);
       string type = "";
       if(code==18){ type = "Buy"; signtype=BUY;}
       else if(code==19){ type = "Buy-Limit"; signtype=BUY_LIMIT;}
       else  if(code==20) {type = "Buy-Stop"; signtype=BUY_STOP;}
       else  if(code==21) {type = "Sell"; signtype=SELL;}
       else  if(code==22) {type = "Sell-Limit"; signtype=SELL_LIMIT;}
       else  if(code==23) {type = "Sell-Stop"; signtype=SELL_STOP;}
       comment = "---Signal Information---\n";
       comment+= "Symbol: " + symbol + "\n";
       comment+= "Type: " + type + "\n";
       comment+= "Entry: " + p + "\n";
       comment+= "Take Profit: " + string(tp) + "\n";
       comment+= "StopLoss: " + string(sl);
       //comment+= "Signal from " +new_signal._comment;
       
       new_signal.type=signtype;
       new_signal.entry=instant? cmp : entry;
       
       return signtype;
   }

   
   void ProcessSignal(CCustomMessage &message)
   {
       string text=ReformText(message.message_text);
       signal_action sig_action=GetAction(text);
       long last=last_signal(message.chat_username);
       long index=-1;
       signal edit_item;
       //Check if message is a reply or edited.
       if((message.reply_to_id!=0) || (message.edited))
       {
           long signal_id=0;
           //bot.SendMessage(bot.user_chat, message.replied_text);
           bot.SendMessage(bot.user_chat, (string)(message.reply_to_id));
           if(message.reply_to_id!=0)
           {
               signal_id=SignalId(message.reply_to_id);
               if(signal_id<100000) signal_id=SignalId(message.replied_text);
           }
           else if(message.edited) signal_id=SignalId(message.message_id);
           index=SignalIndex(signal_id);
           
           if((sig_action!=0) && (index>=0)) {edit_item=signals[index]; ArrayPrint(edit_item.message_ids); ArrayPrint(edit_item.tickets);ProcessAction(edit_item,sig_action,text);}
           
           
           else if(ValidityCheck(text,message))
           {
               //Create signal item.
               Print("Done");
               return;
               //if(sources[a-1]!=message.chat_username)) {ArrayResize(sources,a+1); sources[a]=message.chat_username;}
           }
       }
       
       else if((last>100000) && (sig_action!=0))
       {
           //ArrayPrint(signals);
           string symbol=Instrument(FindW(text,true),text);
           ordertype type=NONE;
           long last=last_signal(message.chat_username); 
           if(StringContains(text,"buy")) type=BUY;
           else if(StringContains(text,"sell")) type=SELL;
           int h=SignalIndex(last);
           if(!h) return;
           edit_item=signals[h];
           ArrayPrint(edit_item.message_ids); ArrayPrint(edit_item.tickets);
           if(symbol!="")
           {
               if(type!=NONE)
               {
                   if((edit_item.symbol==symbol) && (edit_item.type==type)) ProcessAction(edit_item,sig_action,text);
               }
               else
               {
                   if((edit_item.symbol==symbol)) ProcessAction(edit_item,sig_action,text);
               }
           }
           
           //--
       }
       //ArrayPrint(signals);
       //Check validity of signal from text.
       else if(ValidityCheck(text,message))
       {
           //Create signal item.
           Print("Done");
           return;
           //if(sources[a-1]!=message.chat_username)) {ArrayResize(sources,a+1); sources[a]=message.chat_username;}
           //--
       }
   }
   //+------------------------------------------------------------------+
   int               SendScreenShot(const long _chat_id,
                                    const string _symbol,
                                    const string _caption,
                                    const ENUM_TIMEFRAMES _period)
   {
      int result=0;

      long chart_id=ChartOpen(_symbol,_period);
      if(chart_id==0)
         return(ERR_CHART_NOT_FOUND);

      ChartSetInteger(ChartID(),CHART_BRING_TO_TOP,true);

      //--- updates chart
      int wait=60;
      while(--wait>0)
      {
         if(SeriesInfoInteger(_symbol,_period,SERIES_SYNCHRONIZED))
            break;
         Sleep(500);
      }

      //--

      ChartRedraw(chart_id);
      Sleep(500);

      ChartSetInteger(chart_id,CHART_SHOW_GRID,false);

      ChartSetInteger(chart_id,CHART_SHOW_PERIOD_SEP,false);

      string filename=StringFormat("%s%d.gif",_symbol,_period);

      if(FileIsExist(filename))
         FileDelete(filename);
      ChartRedraw(chart_id);

      Sleep(100);

      if(ChartScreenShot(chart_id,filename,800,600,ALIGN_RIGHT))
      {
         
         Sleep(100);
         
         //--- Need for MT4 on weekends !!!
         ChartRedraw(chart_id);
         
         bot.SendChatAction(_chat_id,ACTION_UPLOAD_PHOTO);

         //--- waitng 30 sec for save screenshot
         wait=60;
         while(!FileIsExist(filename) && --wait>0)
            Sleep(500);

         //---
         if(FileIsExist(filename))
         {
            string screen_id;
            result=bot.SendPhoto(screen_id,_chat_id,filename,_caption);
         }
         else
         {
            string mask="Screenshot file '%s' not created.";//--
            PrintFormat(mask,filename);
         }
      }

      ChartClose(chart_id);
      return(result);
   }

   //+------------------------------------------------------------------+
   void              ProcessMessages(void)
   {

      for(int i=0; i<m_chats.Total(); i++)
      {
         CCustomChat *chat=m_chats.GetNodeAtIndex(i);
         
         //Print("e reach me");
         if(!chat.m_new_one.done)
         {
            chat.m_new_one.done=true;
            string text=chat.m_new_one.message_text;
            CCustomMessage message=chat.m_new_one;
            text.Lower();
            static string p_txt="";
            
            if(user_chat=="" && message.chat_type=="private") user_chat=(string)message.chat_id;
            
            if((bot.Name()==message.from_username) || (p_txt==text)) continue;
            p_txt=text;

            //--- start
            if(text=="/start" || text=="/help")
            {
               chat.m_state=0;
               string msg="The bot works with your trading account:\n";
               msg+="/info - get account information\n";
               msg+="/orders - get info on pending orders\n";
               msg+="/positions - get list of open positions\n";

               if(message.chat_type=="private") bot.SendMessage(message.chat_id,msg);
               continue;
            }

            //---
            //---
            else if(text=="/info" || text=="account info")
            {
               string currency=AccountInfoString(ACCOUNT_CURRENCY);
               string msg=StringFormat("%d: %s\n",AccountInfoInteger(ACCOUNT_LOGIN),AccountInfoString(ACCOUNT_SERVER));
               msg+=StringFormat("%s: %.2f %s\n","Balance",AccountInfoDouble(ACCOUNT_BALANCE),currency);
               msg+=StringFormat("%s: %.2f %s\n","Profit",AccountInfoDouble(ACCOUNT_PROFIT),currency);
               if(message.chat_type=="private") bot.SendMessage(message.chat_id,msg);
               continue;
            }

            //---
            else if(text=="/orders" || text=="check orders")
            {
               //--
               string msgs[];
               GetOrders(msgs);
               //--
               for(int i=0; i<ArraySize(msgs); i++) {if(message.chat_type=="private") bot.SendMessage(message.chat_id,msgs[i]); Sleep(1000);}
               continue;
            }

            //---
            else if(text=="/positions" || text=="check positions")
            {
               //--
               string msgs[];
               GetPositions(msgs);
               //--
               for(int i=0; i<ArraySize(msgs); i++) {if(message.chat_type=="private") bot.SendMessage(message.chat_id,msgs[i]); Sleep(1000);} //bot.SendScreenShot(message.chat_id,StringSubstr(msgs[i],0,6),msgs[i],PERIOD_M5); Sleep(1000);}
               continue;
            }

            //---
            else
            {
                ProcessSignal(message);
                continue;
            }
            
        }
      }
   }
};



bool IsMicro()
{
    string suffix;
    for (int s = 0; s < SymbolsTotal(false); s++)
    {
        
        string symbolname = SymbolName(s, false);
        suffix="not_a_micro";
         // Skip non-Forex pairs.
        if ((SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX) && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE)) continue;
        if(StringContains("EURUSD", symbolname) && StringLen(symbolname)>6) suffix=StringSubstr(symbolname, 6,StringLen(symbolname)-6);
        //--
        //Print(suffix);
        if(SymbolInfoDouble("EURUSD" + suffix ,SYMBOL_TRADE_CONTRACT_SIZE)==1000) break;
        else suffix="not_a_micro";
    }
    micro_suffix=suffix;
    return StringContains(_Symbol,suffix);
}


//+------------------------------------------------------------------+
#define EXPERT_NAME     "Telegram Signal Copier"
#define EXPERT_VERSION  "1.00"
#property version       EXPERT_VERSION
#define CAPTION_COLOR   clrWhite
#define LOSS_COLOR      clrOrangeRed

//+------------------------------------------------------------------+
//|   Input parameters                                               |
//+------------------------------------------------------------------+

input group "Telegram Bot Settings"
input ENUM_UPDATE_MODE  InpUpdateMode=UPDATE_NORMAL;//Update Mode
input string            InpToken="7405907024:AAEuBMfqQ6cAYQ_3h1WMyS0z3d1KUzEP3Y4";//Token
input string            InpUserNameFilter="";//Whitelist Usernames
input string            InpUserChat="@goldblues";//Username of Chat for getting Updates.

input group "Activation"
input string license_number="24b12c2028"; //Enter a license code.
input group "Money management"
input double Lots = 0.01; //Lotsize.
input bool MM  = true; // Money Management, if true - position sizing based on stop-loss.
input double Risk = 0.5; // Risk - Risk tolerance in percentage points.
input double FixedBalance = 0; // FixedBalance: If > 0, trade size calc. uses it as balance.
input double MoneyRisk = 0; // MoneyRisk: Risk tolerance in account currency.
input bool UseMoneyInsteadOfPercentage = false; // Use money risk instead of percentage.
input bool UseEquityInsteadOfBalance = true; // Use equity instead of balance.
input double UserMaxLot=30; //Max lot if using auto position sizing.
input double UserMinLot=0.01; //Min lot if using auto position sizing.
input bool no_money_no_trade=false;//Don't use minimum lot if margin is low.
input int Slippage = 3;
input ulong umagic=NULL;

input group "Trading"
input bool mult_pos=true; //Multiple tps opens multiple positions.
input enum_tps use_tp=2; //Takeprofit level if opening single position.
input bool split_risk=true; //Divide risk across positions for same signal.
input bool use_own_sl=false;//Override Stoploss in signal.
input bool use_own_tp=false;//Override TakeProfit in signal.
input bool use_in_absence=true;//Use your own SL or TP when not present in signal.
input bool no_sl=true;//Ignore Signal if stoploss is not available.
input double sl_pips=500; //User Sl in points.
input double tp_pips=800; //User Tp in points.

input bool use_rr_tp=false;//Set User takeprofit in RR.
input double tp_rr=2.5;//Risk to Reward for takeprofit.
input enum_tps be_tp=0; //Takeprofit Level to activate Breakeven.
input bool break_at_rr=true;//Close Partials and set be on RR.
input double be_rr=1.5;//BreakEven RR.

input bool no_stack=false; // Don't stack positions.
input int max_pos=3;//Maximum number of positions to open at a time.
input bool no_sell=false; // Don't Open Sell/Short Positions.
input bool no_buy=false; //Don't Open Buy/Long Positions.
input hedge_type Hedge=0; //Allow Hedging?

input group "Symbols Filter"
input string untradables ="usdzar,usdnok"; //Symbols not to trade (multiple symbols separated by commas.
input bool trade_specific=false; //Trade only specified symbols?
input string tradables="gold, gbpusd, eurusd, gbpjpy, gbpcad";//Specific symbols to trade.
input string symbol_map="bitcoin-btcusd,litecoin-ltcusd,ripple-xrpusd,";// Symbol Map. (alternative name-symbolname). Multiple mappings separated by commas.

input group"News Settings"
input bool avoid_events=false; //Pause trading during news events.
input curr_enum curr = All; // Currency's news event to avoid.
input bool usd_only = false; // Avoid USD news.
input bool OnlySpecificNews = true; // Avoid only specific news.
input string NewsNames = "Nonfarm Payrolls,GDP,CPI,Press Conference,PMI,PPI,Interest Rate Decision"; // Title of special news events to avoid. Multiple titles separated by comma.
input string VoidList = "crude oil,speech"; //Title of news events to ignore. Multiple titles separated by comma.
input int OpenSecs = 2; // Minutes before and after the news event to pause trading activities. 
input close_positions close_news=leave_all;// Minutes before news;

input group "Sessions"
input bool trade_assian=false; //Trade during Assian session.
input bool trade_london=true; //Trade during London session.
input bool trade_newyork=true; //Trade during New York session.
input datetime t_starting=D'1970.01.01 00:00:00'; //Time to start trading (GMT).
input datetime t_ending=D'1970.01.01 23:59:59'; //Time to stop trading (GMT).

input group "Trading Days"
input bool trade_mondays=true; //Trade on mondays.
input bool trade_teusdays=true; //Trade on Teusdays.
input bool trade_wednesdays=true; //Trade on Wednesdays.
input bool trade_thursdays=true; //Trade on Thursdays.
input bool trade_fridays=true; //Trade on Fridays.
input bool trade_saturdays=true; //Trade on Saturdays.
input bool trade_sundays=true; //Trade on Sundays.

input group "Risk Management"
input close_positions close_target=leave_all; //When target or max drawdown is reached.
input double max_profit=100; //Target profit in percentage to end day's trading.
input double max_dd=45; //Max drawdown to pause day's trading.
input bool use_amount=false; //Use Money targets.
input double profit_amount=0; //Money profit to end day's trading.
input double loss_amount=0; //Money loss to end day's trading.
input bool take_equity=true; //Drawdown equity based. 
input close_positions close_weekend=leave_all;//Just before Weekend.

input group "Timer"
input bool ShowTimer = true; // Always Show Timer.
input int FontSize = 18;
input string Font = "Arial";
input color FontColor = clrDarkOrange;
input ENUM_BASE_CORNER Corner = CORNER_LEFT_UPPER;
input int X_Distance = 10; // X-axis distance from the chart corLTner.
input int Y_Distance = 130; // Y-axis distance from the chart corLTner.

// Global variables:
//---
CComment       display_comment;
CMyBot         bot;
ENUM_RUN_MODE  run_mode;
datetime       time_check;
int            web_error;
int            init_error;
string         photo_id=NULL;
string tradable_pairs, non_tradable_pairs, pair_key[], pair_attribute[];

bool ev_high = true;// Trade high importance news.

bool ev_moderate = true; //Trade medium importance news.

bool ev_low = false; //Trade low importance news.

bool freqFilter = true; //Filter off weekly releases.

bool doublenews = false; // Don't avoid trade if there are news releases for both currencies in a pair at the same time.

string news_names[], avoid_list[], curren[2], m_country, m_currency, Commentary="";

int news_time = 0,total_pos=0;

bool CanTrade = true, Activated=false;

bool Terminal_Trade_Allowed = true;

int corLT, corTG, corLG;
double BE,TP;
bool pos_check=false;
string Names="", pre_text;
bool micro;

NewsData newsarray[];
signal signals[];
string sources[], micro_suffix;

CNews news;

//Globals;

datetime Exp_date=0;
// For tick value adjustment:

string AccountCurrency = "";

string ProfitCurrency = "";

string BaseCurrency = "";

ENUM_SYMBOL_CALC_MODE CalcMode;

string ReferencePair = NULL;

bool ReferenceSymbolMode;
order order_list[];
ulong Magic;

// Main trading objects:

CTrade *Trade; 

CPositionInfo PositionInfo;

CSymbolInfo SymbolInfo;
CAccountInfo AccountInfo;


int OnInit()
{
    OnBotInit();
    OrderInit();
    //days_lookback = PeriodSeconds(PERIOD_CURRENT)/PeriodSeconds(PERIOD_M15);
    NewsInit();
    init_filters();
    
    micro=IsMicro();
    
    string msg=license_number;
    if((!Activated) && (!DecodeLicense(msg))) Alert("Please enter a valid licence code.");
    else {if(msg!=license_number) {Alert(msg);}; return (INIT_SUCCEEDED);}
    return (INIT_SUCCEEDED);
}



void OnDeinit(const int reason)
{ 
    OnBotDeinit(reason);
    OrderDeinit(reason);
    NewsDeinit(reason);
    return;
}


//+------------------------------------------------------------------+
//| Updates text about time left to news release or passed after news release.      |
//+------------------------------------------------------------------+

void OnTimer()
{
    OnBotTimer();
    //if(Exp_date<TimeCurrent()) {Alert("License Expired!"); ExpertRemove();}
    if(((news_time + (OpenSecs*60))<=(int)TimeLocal()) && news_time!=0) { news_time= GetNews();}
    if(news_time==0)
    {
        if(IsNewsBar()) news_time=GetNews();
        else return;
    }

    string text;
    int difference = (int)TimeLocal() - news_time;

    if(difference <= (OpenSecs*60)) text = "Next pause in: " + TimeCountString(-difference);

    else if(difference >= (OpenSecs*60)) text = "All caught up!";

    if((news_time == 0) || (difference >= (OpenSecs*60))) {ObjectSetString(0, "NewsTradingInfo", OBJPROP_TEXT, "Trade Safe!");}

    ObjectSetString(0, "News trading pa>99>I'llTime of>it", OBJPROP_TEXT, text + ".");
    ObjectSetInteger(0, "NewsTradingTime", OBJPROP_FONTSIZE, FontSize);
    ObjectSetString(0, "NewsTradingTime", OBJPROP_FONT, Font);
    ObjectSetInteger(0, "NewsTradingTime", OBJPROP_COLOR, FontColor);

    ObjectSetInteger(0, "NewsTradingInfo", OBJPROP_FONTSIZE, FontSize);
    ObjectSetString(0, "NewsTradingInfo", OBJPROP_FONT, Font);
    ObjectSetInteger(0, "NewsTradingInfo", OBJPROP_COLOR, FontColor);
    
    if(news_time!=0)
    {
        ObjectSetString(0, "NewsTradingInfo", OBJPROP_TEXT, Commentary);
    }
    ChartRedraw(0);
    OnTick();
}

void OnChartEvent(const int id,const long &lparam,const double &dparam,const string &sparam)
{
   display_comment.OnChartEvent(id,lparam,dparam,sparam);
}

void OnTick()
{
    //--
    ControlOrders();
    BreakEven();
}


int OnBotInit()
{

//---
   run_mode=GetRunMode();

//--- stop working in tester
   if(run_mode!=RUN_LIVE)
   {
      PrintError(ERR_RUN_LIMITATION,LANGUAGE_EN);
      return(INIT_FAILED);
   }

   int y=40;
   if(ChartGetInteger(0,CHART_SHOW_ONE_CLICK))
      y=120;
   display_comment.Create("myPanel",20,y);
   display_comment.SetColor(clrDimGray,clrBlack,220);

//--- set token
   init_error=bot.Token(InpToken);

//--- set filter
   bot.UserNameFilter(InpUserNameFilter);
   
//--- set user chat;
   bot.user_chat=InpUserChat;

//--- set timer
   int timer_ms=3000;
   switch(InpUpdateMode)
   {
   case UPDATE_FAST:
      timer_ms=1000;
      break;
   case UPDATE_NORMAL:
      timer_ms=2000;
      break;
   case UPDATE_SLOW:
      timer_ms=3000;
      break;
   default:
      timer_ms=3000;
      break;
   };
   EventSetMillisecondTimer(timer_ms);
   OnTimer();
//--- done
   return(INIT_SUCCEEDED);
}
//+------------------------------------------------------------------+
//|   OnDeinit                                                       |
//+------------------------------------------------------------------+
void OnBotDeinit(const int reason)
{
//---
   if(reason==REASON_CLOSE ||
         reason==REASON_PROGRAM ||
         reason==REASON_PARAMETERS ||
         reason==REASON_REMOVE ||
         reason==REASON_RECOMPILE ||
         reason==REASON_ACCOUNT ||
         reason==REASON_INITFAILED)
   {
      time_check=0;
      display_comment.Destroy();
   }
//---
   EventKillTimer();
   ChartRedraw();
}

void OnBotTimer()
{

//--- show init error
   if(init_error!=0)
   {
      //--- show error on display
      CustomInfo info;
      GetCustomInfo(info,init_error,LANGUAGE_EN);

      //---
      display_comment.Clear();
      display_comment.SetText(0,StringFormat("%s v.%s",EXPERT_NAME,EXPERT_VERSION),CAPTION_COLOR);
      display_comment.SetText(1,info.text1, LOSS_COLOR);
      if(info.text2!="")
         display_comment.SetText(2,info.text2,LOSS_COLOR);
      display_comment.Show();

      return;
   }

//--- show web error
   if(run_mode==RUN_LIVE)
   {

      //--- check bot registration
      if(time_check<TimeLocal()-PeriodSeconds(PERIOD_H1))
      {
         time_check=TimeLocal();
         if(TerminalInfoInteger(TERMINAL_CONNECTED))
         {
            //---
            web_error=bot.GetMe();
            if(web_error!=0)
            {
               //---
               if(web_error==ERR_NOT_ACTIVE)
               {
                  time_check=TimeCurrent()-PeriodSeconds(PERIOD_H1)+300;
               }
               //---
               else
               {
                  time_check=TimeCurrent()-PeriodSeconds(PERIOD_H1)+5;
               }
            }
         }
         else
         {
            web_error=ERR_NOT_CONNECTED;
            time_check=0;
         }
      }

      //--- show error
      if(web_error!=0)
      {
         display_comment.Clear();
         display_comment.SetText(0,StringFormat("%s v.%s",EXPERT_NAME,EXPERT_VERSION),CAPTION_COLOR);

         if(
#ifdef __MQL4__ web_error==ERR_FUNCTION_NOT_CONFIRMED #endif
#ifdef __MQL5__ web_error==ERR_FUNCTION_NOT_ALLOWED #endif
         )
         {
            time_check=0;

            CustomInfo info= {0};
            GetCustomInfo(info,web_error,LANGUAGE_EN);
            display_comment.SetText(1,info.text1,LOSS_COLOR);
            display_comment.SetText(2,info.text2,LOSS_COLOR);
         }
         else
            display_comment.SetText(1,GetErrorDescription(web_error,LANGUAGE_EN),LOSS_COLOR);

         display_comment.Show();
         return;
      }
   }

//---
   bot.GetUpdates();

//---
   if(run_mode==RUN_LIVE)
   {
      display_comment.Clear();
      display_comment.SetText(0,StringFormat("%s v.%s",EXPERT_NAME,EXPERT_VERSION),CAPTION_COLOR);
      display_comment.SetText(1,StringFormat("%s: %s","Bot Name",bot.Name()),CAPTION_COLOR);
      display_comment.SetText(2,StringFormat("%s: %d","Chats",bot.ChatsTotal()),CAPTION_COLOR);
      display_comment.Show();
   }

//---
   bot.ProcessMessages();
}
//+------------------------------------------------------------------+
//|   GetCustomInfo                                                  |
//+------------------------------------------------------------------+
void GetCustomInfo(CustomInfo &info,
                   const int _error_code,
                   const ENUM_LANGUAGES _lang)
{
   switch(_error_code)
   {
#ifdef __MQL5__
   case ERR_FUNCTION_NOT_ALLOWED:
      info.text1 = "The URL does not allowed for WebRequest";//Этого URL нет в списке для WebRequest.";
      info.text2 = TELEGRAM_BASE_URL;
      break;
#endif
#ifdef __MQL4__
   case ERR_FUNCTION_NOT_CONFIRMED:
      info.text1 = "The URL does not allowed for WebRequest";//Этого URL нет в списке для WebRequest.";
      info.text2 = TELEGRAM_BASE_URL;
      break;
#endif

   case ERR_TOKEN_ISEMPTY:
      info.text1 = "The 'Token' parameter is empty.";//Параметр 'Token' пуст.";
      info.text2 = "Please fill this parameter.";//Пожалуйста задайте значение для этого параметра.";
      break;
   }

}
//+------------------------------------------------------------------+


void NewsInit()
{
    if(!StringContains(NewsNames, ",")) {ArrayResize(news_names, 1); news_names[0] = NewsNames;}

    else ArrayResize(news_names, StringSplit(NewsNames, StringGetCharacter(",", 0), news_names));

    
    if(!StringContains(VoidList, ",")) {ArrayResize(avoid_list, 1); avoid_list[0] = VoidList;}

    else ArrayResize(avoid_list, StringSplit(VoidList, StringGetCharacter(",", 0), avoid_list));

   
    corTG= (int)TimeTradeServer() - (int)TimeGMT();


    if(NewsNames == "" && OnlySpecificNews) {Alert("You opted to trade only specific news. Please specify the title(s)"); ExpertRemove();}

    curren[0] = SymbolInfoString( _Symbol,SYMBOL_CURRENCY_PROFIT);

    curren[1] = SymbolInfoString( _Symbol,SYMBOL_CURRENCY_BASE);

    

    corLT = (int)TimeLocal() - (int)TimeTradeServer();
    corLG = corTG + corLT;

    

    news_time= GetNews();
    
    if(ShowTimer)
    {
        ObjectCreate(ChartID(), "NewsTradingInfo", OBJ_LABEL, 0, 0, 0);

        ObjectSetInteger(ChartID(), "NewsTradingInfo", OBJPROP_CORNER, Corner);

        ObjectSetInteger(ChartID(), "NewsTradingInfo", OBJPROP_XDISTANCE, X_Distance);

        ObjectSetInteger(ChartID(), "NewsTradingInfo", OBJPROP_YDISTANCE, Y_Distance);

        ObjectSetInteger(ChartID(), "NewsTradingInfo", OBJPROP_SELECTABLE, true);
       
       
        ObjectCreate(ChartID(), "NewsTradingTime", OBJ_LABEL, 0, 0, 0);

        ObjectSetInteger(ChartID(), "NewsTradingTime", OBJPROP_CORNER, Corner);

        ObjectSetInteger(ChartID(), "NewsTradingTime", OBJPROP_XDISTANCE, X_Distance);

        ObjectSetInteger(ChartID(), "NewsTradingTime", OBJPROP_YDISTANCE, Y_Distance + (FontSize + 15));

        ObjectSetInteger(ChartID(), "NewsTradingTime", OBJPROP_SELECTABLE, true);

        EventSetMillisecondTimer(900); // For smooth updates.
    }
}

//+------------------------------------------------------------------+

//| Deletes graphical object if needed.                              |

//+------------------------------------------------------------------+

void NewsDeinit(const int reason)

{

    ObjectDelete(ChartID(), "NewsTradingTime");

    ObjectDelete(ChartID(), "NewsTradingInfo");

}

//+------------------------------------------------------------------+
//| Format time distance from the number of seconds to normal string |
//| of years, days, hours, minutes, and seconds.                     |
//| secs - number of seconds                                            |
//| Returns: formatted string.                                       |
//+------------------------------------------------------------------+

string TimeCountString(int secs)
{
    if (secs == 0) return("0secs");

    string t_str = "";

    int yr = 0;

    int d = 0;

    int hr = 0;

    int mins = 0;

    yr = secs / 31536000;

    secs-= yr * 31536000;

    d = secs / 86400;

    secs-= d * 86400;

    hr = secs / 3600;

    secs-= hr * 3600;

    mins = secs / 60;

    secs-= mins * 60;

    if (yr) t_str += IntegerToString(yr) + "yr";

    if (yr > 1) t_str += "s";


    if (d) t_str += " " + IntegerToString(d) + "day";

    if (d > 1) t_str += "s";

    if (hr) t_str += " " + IntegerToString(hr) + "hr";

    if (hr > 1) t_str += "s";

    if (mins) t_str += " " + IntegerToString(mins) + "min";

    if (mins > 1) t_str += "s";


    if (secs) t_str += " " + IntegerToString(secs) + "sec";

    if (secs > 1) t_str += "s";

    StringTrimLeft(t_str);

    return t_str;
}

bool IsNewsBar()
{
    static datetime saved_candle_time=0;

    if(iTime(Symbol(), PERIOD_D1,0) == saved_candle_time) return false;
    else saved_candle_time = iTime(Symbol(),PERIOD_D1,0);

    return true;
} 

int GetNews()
  {

      int time = 0;
      static datetime last_update=TimeLocal();
  
      if(last_update-(PeriodSeconds(PERIOD_D1)*14) < TimeLocal())
      {
         ArrayFree(newsarray);
   
         int k = 0,s=0;
   
         //Print("News Data Loaded.");
   
         for(int l=0; l<2; l++)
   
         {
   
            m_currency = curren[l];
   
   
   
            int x  = news.update(60,true);
   
         
   
         // Detect news
   
            bool HaveNews           = x > 0 ? true : false;
   
   
   
          // Stored news data to struct
   
            if(!HaveNews) continue;
   
             
   
            ArrayResize(newsarray, (x+k));
   
            
   
             for(int i = 0; i < x; i++)
   
             {
   
                if((ev_high && news.event[i].importance == CALENDAR_IMPORTANCE_HIGH) || (ev_moderate && news.event[i].importance == CALENDAR_IMPORTANCE_MODERATE) || (ev_low && news.event[i].importance == CALENDAR_IMPORTANCE_LOW))
   
                {
   
                   if(freqFilter && news.event[i].frequency==CALENDAR_FREQUENCY_WEEK) continue;
   
                   if(news.CountryIdToCurrency((ENUM_COUNTRY_ID)news.event[i].country_id) != m_currency) continue;
   
                   newsarray[k].NewsCurrency    = m_currency;
   
                   newsarray[k].NewsName     = news.eventname[i];
   
                   newsarray[k].Importance = news.event[i].importance;
   
                   newsarray[k].ReleaseTime  = news.event[i].time;
   
                   k++;
   
                  }
   
               }
   
           }
   
           ArrayResize(newsarray, k);
          
         //Re-arrange array items by time.
   
         for(int j = 0; j<ArraySize(newsarray); j++)
   
         {
   
             for(int i = 0; i < ArraySize(newsarray);i++)
   
             {
   
                 NewsData swap;
   
                 swap.ReleaseTime = newsarray[j].ReleaseTime ;
   
                 
   
                 swap.Importance = newsarray[j].Importance;
   
                 swap.NewsCurrency = newsarray[j].NewsCurrency;
   
                 swap.NewsName = newsarray[j].NewsName;
   
                 if(newsarray[j].ReleaseTime < newsarray[i].ReleaseTime)
   
                 {
   
                     newsarray[j].ReleaseTime = newsarray[i].ReleaseTime ;
   
                     
   
                     newsarray[j].Importance = newsarray[i].Importance;
   
                     newsarray[j].NewsCurrency = newsarray[i].NewsCurrency;
   
                     newsarray[j].NewsName = newsarray[i].NewsName;
   
                     newsarray[i].ReleaseTime = swap.ReleaseTime;
   
                     
   
                     newsarray[i].Importance = swap.Importance;
   
                     newsarray[i].NewsCurrency = swap.NewsCurrency;
   
                     newsarray[i].NewsName = swap.NewsName;
                 }
   
             }
             last_update=newsarray[ArraySize(newsarray)-2].ReleaseTime + corLT;
             //Print("Yay!");
         }
     }
     // ////ArrayPrint(newsarray);
      

      if(ArraySize(newsarray) > 1)

      {

         for(int i = 1; i < ArraySize(newsarray); i++)

         {

             datetime BeforeReleaseTime       = newsarray[i-1].ReleaseTime - PeriodSeconds(PERIOD_MN1);

            // Load news into news_time for trading with preference checks.

            if((TimeLocal() > BeforeReleaseTime) && (TimeTradeServer() < newsarray[i-1].ReleaseTime))

             {

                 bool Special=false,Specific=false;

                 string txt = newsarray[i-1].NewsName;

                 StringToLower(txt);

                 for(int n = 0; n<ArraySize(news_names); n++)
                 {
                     string match = news_names[n];
                     StringToLower(match);
                     StringReplace(match, "nfp", "nonfarm payrolls");
                     StringReplace(match, "funds rate", "interest rate");

                     if(StringContains(txt, match)) {Specific = true; break;}

                 }

                 bool avoid = false;

                 for(int n = 0; n<ArraySize(avoid_list); n++)

                 {

                     string match = avoid_list[n];

                     StringToLower(match);

                     if(StringContains(txt, match)) {avoid = true; break;}

                 }

                 if(avoid) continue;

                 if((usd_only && newsarray[i-1].NewsCurrency!="USD") || ((curr == Base) && (newsarray[i-1].NewsCurrency != curren[1])) || ((curr == Profit) && (newsarray[i-1].NewsCurrency != curren[0]))) continue;

                 if((OnlySpecificNews) && (!Specific))  continue;

                  Commentary =   "News Alert! " + newsarray[i-1].NewsName +" (" + newsarray[i-1].NewsCurrency + ").";

                  time = (int)newsarray[i-1].ReleaseTime + corLT;
                
                  break;

             }

         }

      }
      return time;
  }

bool StopTrading()
{
    static bool done=false,stat=false;
    MqlDateTime time={};
    datetime start=(iTime(_Symbol,PERIOD_D1,0))+(t_starting), stop=(iTime(_Symbol,PERIOD_D1,0))+(t_ending);
    datetime now =TimeGMT();
    if(done) return true;
    if(!TimeToStruct(now,time)) Alert("TimeToStruct Error!");
    if((avoid_events) && ((news_time>TimeLocal() && ((news_time-TimeLocal())<=(OpenSecs*60))) || (news_time<TimeLocal() && ((TimeLocal()-news_time)<=(OpenSecs*60))))) {if(pos_check) TerminatePositions(close_news); return true;}
    if((now>stop) || (now<start)) {stat=false; return true;}
    if((!trade_assian) && ((time.hour>=21) || (time.hour<7))) return true;
    if((!trade_london) && (time.hour>=7) && (time.hour<12)) return true;
    if((!trade_newyork) && (time.hour>=12) && (time.hour<16)) return true;
    if((!trade_sundays) && (time.day==0)) return true;
    if((!trade_mondays) && (time.day==1)) return true;
    if((!trade_teusdays) && (time.day==2)) return true;
    if((!trade_wednesdays) && (time.day==3)) return true;
    if((!trade_thursdays) && (time.day==4)) return true;
    if((!trade_fridays) && (time.day==5)) return true;
    if((!trade_saturdays) && (time.day==6)) return true;
    if(TargetReached()) {done=TerminatePositions(close_target); return true;}
    if((close_weekend<=3) && (time.day==5) && (time.hour>=20)) {done=TerminatePositions(close_weekend); Alert("Trading week concluded!"); return true;}
    //stat=stat || (!IsNewsBar());
    if(stat) return false;
    if((time.day==5) && (time.hour>=20)) stat=SendSignalStats(5);
    if(time.hour>=23) stat=SendSignalStats(0);
    
    return false;
}

bool TerminatePositions(close_positions close_type=leave_all)
{
    if(close_type==leave_all) return true;
    double sl_points=0;
    double profit=0;
    ulong ticket;
    bool done=false;
    int total=PositionsTotal(), count=0;
    string text=" ";
    if(close_type==close_all) text="Closing all positions!";
    else if(close_type==close_profit) text="Closing all profitable positions!";
    else if(close_type==close_loss) text="Closing all losing positions!";
    else if(close_type==leave_be) text="Closing all positions not on breakeven!";
    for(int p=total-1; p>=0; p--)
    {
        //filter.
        /**
        if(PositionGetInteger(POSITION_MAGIC) != Magic) continue;
        if (PositionGetSymbol(p) != Symbol()) continue;
        **/
        PositionInfo.SelectByIndex(p);
        ticket=PositionInfo.Ticket();
        double open=PositionInfo.PriceOpen();
        double sl=PositionInfo.StopLoss();
        bool buy=PositionInfo.PositionType() == POSITION_TYPE_BUY;
        sl_points=buy? open-sl : sl-open;
        profit=PositionInfo.Profit();
        
        if(close_type==close_all) {count=count+1; Trade.PositionClose(ticket);}
        if((close_type==close_profit) && (profit>0)) {count=count+1; done=(done) || Trade.PositionClose(ticket);}
        if((close_type==close_loss) && (profit<0)) {count=count+1; done=(done) || Trade.PositionClose(ticket);}
        if((close_type==leave_be) && (sl_points<0)) {count=count+1; done=(done) || Trade.PositionClose(ticket);}
    }
    if(count>0) Alert(text);
    else done=true;
    return done;
}

bool ClearOpposite(bool buy=true)
{
    CPositionInfo posinfo;
    CTrade pos;
    bool clear=true;
    if(Hedge==allow_hedge) return true;
    for(int i=0; i<3; i++)
    {
        for(int i=PositionsTotal()-1; i>=0; i--)
        {
            posinfo.SelectByIndex(i);
            if(posinfo.Magic()!=Magic) continue;
            if(posinfo.Symbol()!=_Symbol) continue;
            if(buy && (posinfo.PositionType()== POSITION_TYPE_SELL))
            {
                if(Hedge==close_opposite) {Print("Avoiding hedge!"); clear=pos.PositionClose(posinfo.Ticket());}
                else return false;
            }
            else if((!buy) && (posinfo.PositionType()== POSITION_TYPE_BUY))
            {
                if(Hedge==close_opposite) {Print("Avoiding hedge!"); clear=pos.PositionClose(posinfo.Ticket());}
                else return false;
            }
        }
    }
    return clear;
}

bool TargetReached()
{
    static double target_profit=0, max_loss=0;
    double balance=AccountInfoDouble(ACCOUNT_BALANCE);
    double equity=AccountInfoDouble(ACCOUNT_EQUITY);
    static datetime prev_day=0;//iTime(Symbol(), PERIOD_D1,0);
    static bool done=false;
    if(prev_day!=iTime(Symbol(), PERIOD_D1,0))
    {
        prev_day=iTime(Symbol(), PERIOD_D1,0);
        done=false;
        target_profit=use_amount? profit_amount + balance : ((max_profit/100) * balance) + balance;
        max_loss=use_amount? balance-loss_amount : balance-((max_dd/100) * balance);
    }
    if(done) return true;
    //if(target_profit==max_loss) return false;
    if((take_equity && (equity>target_profit)) || ((!take_equity) && (balance>target_profit))) {Alert("Target profit hit @ ", target_profit); done= true;}
    else if((take_equity && (equity<=max_loss)) || ((!take_equity) && (balance<=max_loss))) {Alert("Max loss hit @ ", max_loss); done= true;}
    return false;
}

void OrderInit()
{
    if((Magic==NULL) && (umagic!=NULL)) Magic = umagic;

    MathSrand((int)GetTickCount()); int s1 = MathRand(); MathSrand((int)TimeLocal()); int s2 =  MathRand();

    if(Magic==NULL) { Magic = fabs(s1 - s2); Print(__FILE__," ",  _Symbol, ": Magic set as ", Magic, ". Please make sure it is not used for any other EA on this terminal."); }
    
    double min_lot = SymbolInfoDouble( _Symbol, SYMBOL_VOLUME_MIN);

    if ((Lots < min_lot) && (!MM)) {Alert("Lots should be not less than: ", DoubleToString(min_lot, 2), "."); return;}
    if(sl_pips==0 && MM && (use_in_absence || use_own_sl)) Alert("Position size can't be calculated without stoploss; Manual lots will be used. when necessary ");

    Trade = new CTrade;

    Trade.SetDeviationInPoints(Slippage);

    Trade.SetExpertMagicNumber(Magic);

}

void OrderDeinit(const int reason){delete Trade;}

void  CreateOrder(signal &new_signal)
{
    ArrayResize(order_list, ArraySize(order_list)+1);
    static int last_index=0;
    string return_text;
    int i = ArraySize(order_list)-1;
    order_list[i].type=new_signal.type;
    bool buy=((int)(new_signal.type))%2!=0;
    if((no_buy && buy) || (no_sell && (!buy))) return;
    
    if((new_signal.tp==0 && use_in_absence && (!use_rr_tp)) || use_own_tp)  new_signal.tp=(buy)? (new_signal.entry+(tp_pips*m_point(new_signal.symbol))) : (new_signal.entry-(tp_pips*m_point(new_signal.symbol))); 
    if((new_signal.sl==0 && use_in_absence) || use_own_sl)  new_signal.sl=(buy)? (new_signal.entry-(sl_pips*m_point(new_signal.symbol))) : (new_signal.entry+(sl_pips*m_point(new_signal.symbol)));
    
    /**
    if(StringContains(new_signal.symbol,"JPY"))
    {
        if((new_signal.tp==0 && use_in_absence && (!use_rr_tp)) || use_own_tp)  new_signal.tp=(buy)? (new_signal.entry+(tp_pips*100*m_point(new_signal.symbol))) : (new_signal.entry-(tp_pips*100*m_point(new_signal.symbol))); 
        if((new_signal.sl==0 && use_in_absence) || use_own_sl)  new_signal.sl=(buy)? (new_signal.entry-(sl_pips*100*m_point(new_signal.symbol))) : (new_signal.entry+(sl_pips*100*m_point(new_signal.symbol)));
    }
    **/
    
    if(use_rr_tp && ((use_in_absence && new_signal.tp==0) || use_own_tp)) new_signal.tp=buy? (new_signal.entry+(fabs(new_signal.entry-new_signal.sl)*tp_rr)) : (new_signal.entry-(fabs(new_signal.entry-new_signal.sl)*tp_rr));
    
    //if(remove_sl) new_signal.sl=0;
    if(total_pos>=max_pos) {Alert("Position not opened; Position limit reached!"); Print("Total open positions: ",total_pos,"."); return;}
    if(StringContains(non_tradable_pairs,new_signal.symbol) || (!StringContains(tradable_pairs,new_signal.symbol)))
    {
        return_text="Order not created! Forbidden Symbol!";
        bot.SendMessage(bot.user_chat,return_text);
        return;
    }
    order_list[i].setup_time=TimeTradeServer();
    order_list[i].comment= (string)(new_signal.signal_id) +" By "+new_signal._comment;
    order_list[i].symbol=new_signal.symbol;
    order_list[i].index=last_index=last_index+1;
    order_list[i].type=new_signal.type;
    order_list[i].limit_price=new_signal.entry;
    order_list[i].price=new_signal.entry;
    order_list[i].stoploss=new_signal.sl;
    order_list[i].take_profit=new_signal.tp;
    if((use_tp!=2) && ArraySize(new_signal.tps)>((int)(use_tp))) order_list[i].take_profit=new_signal.tps[(int)(use_tp)];
    order_list[i].signal_id=new_signal.signal_id;
    order_list[i].active=false;
    order_list[i].expiration=(int)(TimeTradeServer()+((2)*PeriodSeconds(PERIOD_MN1)));
    ArrayResize(order_list[i].tps,ArraySize(new_signal.tps));
    for(int j=0; j<ArraySize(new_signal.tps); j++) {order_list[i].tps[j]=new_signal.tps[j];}
    //order_list[i].tps=new_signal.tps;
    
    return_text= new_signal.symbol + " " +EnumToString(new_signal.type) + " Order created." +"\n" + "Signal Id: "+ order_list[i].comment;
    if(StringContains(non_tradable_pairs,new_signal.symbol) || (!StringContains(tradable_pairs,new_signal.symbol))) return_text="Order not created! Forbidden Symbol!";
    bot.SendMessage(bot.user_chat,return_text);
    
    //ArrayPrint(order_list);
}

bool DecodeLicense(string &msg)
{
    int day=(int)(StringSubstr(msg,0,2));
    int month=(int)(StringSubstr(msg,3,2));
    int year=(int)(StringSubstr(msg,6,4));
    datetime exp= (datetime)(((year-1970)*365*24*60*60) + ((month*30*24*60*60)+(day*24*60*60)));
    //Print(TimeCurrent(), "\n ", exp);
    if(exp<=TimeCurrent()) {msg = "Invalid or expired license!"; return false;}
    else
    {
        Exp_date=exp;
        msg="Successful activation! License valid for " + TimeCountString(((int)(exp-TimeCurrent()))) + ".";
        Activated=true;
    }
    return true;
}

void ControlOrders()
{
    int time = (int)TimeTradeServer();
    double Ask = SymbolInfoDouble( _Symbol, SYMBOL_ASK);

    if(ArraySize(order_list)>0)
    {
        for (int i = 0; i < ArraySize(order_list); i++)

        {
            order order = order_list[i];
            if(order.active) continue;
            TradeCase(order);
            if((order.expiration<time) && (!order.active))
            {
                if(OrderPop(order.index)) {Print(EnumToString(order.type), " Deleted!");}
            }
        }

    }

}

bool OrderPop(int order_index)
{
    order next_order;
    int index =orderindex(order_index);

    //ArrayPrint(order_list);

    int total = ArraySize(order_list);
    //if(order_index>total-1) {Print("ERROR OCCURRED!"); return false;}
    if(index==-1) return false;
    for(int i=index; i<total-1; i++)
    {
        //if(i==total-2) break;
        next_order=order_list[i+1];
        order_list[i]=next_order;
    }

    ArrayResize(order_list,total-1);

    //ArrayPrint(order_list);

    return true;

}

//+------------------------------------------------------------------+

//| Generic buy.                                                     |

//+------------------------------------------------------------------+

bool m_Buy(order &position)

{
    string symbol= position.symbol;
    if(!ClearOpposite(true)) return false;
    Print("Opening Buy.");

    double Ask = SymbolInfoDouble( symbol, SYMBOL_ASK);
    // 
    double new_sl = position.stoploss, new_tp = position.take_profit;
    if((no_sl && (new_sl==0)) || no_buy) return false;

    long spread = (long)(SymbolInfoInteger(_Symbol, SYMBOL_SPREAD)*SymbolInfoDouble(_Symbol, SYMBOL_POINT));
    if (spread >= fabs(new_sl-Ask))
    {
        Print( symbol, ": Spread fuse prevents positions from opening. Current spread: ", spread/SymbolInfoDouble(_Symbol, SYMBOL_POINT), " points.");
        return true;
    }

    double lots=LotsOptimized(symbol,ORDER_TYPE_BUY,0,fabs(position.limit_price-position.stoploss));
    if(lots==0) {bot.SendMessage(bot.user_chat,"Insufficient margin for opening position"); return false;}
    int n_tp=ArraySize(position.tps);
    if((n_tp>1) && mult_pos)
    {
        int i=n_tp;
        if(split_risk) i=max_divisor(lots,n_tp,symbol);
        while(--i>=0 && position.tps[i]>0) {new_tp=position.tps[i]; Trade.PositionOpen(symbol, ORDER_TYPE_BUY, lots, Ask, new_sl, new_tp, position.comment); bot.SendMessage(bot.user_chat,Trade.ResultRetcodeDescription());}
        return true;
    }
    
    if(Trade.PositionOpen( symbol, ORDER_TYPE_BUY, lots, Ask, new_sl, new_tp, position.comment)) 
    {
        bot.SendMessage(bot.user_chat,"Buy order executed.");
        Print("Buy Order executed.");
        int i=SignalIndex(position.signal_id);
        if(i>=0)
        {
            int d=ArraySize(signals[i].tickets)-1;
            if(d>=0) {ArrayResize(signals[i].tickets, d+2); signals[i].tickets[d]=Trade.RequestPosition();}
        }
        return true;
    }
    bot.SendMessage(bot.user_chat,Trade.ResultRetcodeDescription());


    return false;

}

//+------------------------------------------------------------------+

//| Generic sell.                                                    |

//+------------------------------------------------------------------+

bool m_Sell(order &position)
{
    string symbol= position.symbol;
    //--
    if(!ClearOpposite(false)) return false;
    Print("Opening Sell.");
    
    double Bid = SymbolInfoDouble( symbol, SYMBOL_BID);

    // Bid and Ask are swapped to preserve the probabilities and decrease/increase profit/loss size.

    double new_sl = position.stoploss, new_tp = position.take_profit;
    if((no_sl && (new_sl==0)) || no_sell) return false;

    long spread = (long)(SymbolInfoInteger( symbol, SYMBOL_SPREAD)*SymbolInfoDouble( symbol, SYMBOL_POINT));
    if (spread >= fabs(new_sl-Bid))
    {
        Print( symbol, ": Spread fuse prevents positions from opening. Current spread: ", spread/SymbolInfoDouble( symbol, SYMBOL_POINT), " points.");
        return true;
    }

    double lots=LotsOptimized(symbol,ORDER_TYPE_SELL,0,fabs(position.limit_price-position.stoploss));
    if(lots==0) {bot.SendMessage(bot.user_chat,"Insufficient margin for opening position"); return false;}
    int n_tp=ArraySize(position.tps);
    if((n_tp>1) && mult_pos)
    {
        int i=n_tp;
        if(split_risk) i=max_divisor(lots,n_tp,symbol);
        while(--i>=0 && position.tps[i]>0) {new_tp=position.tps[i]; Trade.PositionOpen(symbol, ORDER_TYPE_SELL, lots, Bid, new_sl, new_tp, position.comment); bot.SendMessage(bot.user_chat,Trade.ResultRetcodeDescription());}
        return true;
    }

    if(Trade.PositionOpen( symbol, ORDER_TYPE_SELL, lots, Bid, new_sl, new_tp, position.comment))  
    {
        bot.SendMessage(bot.user_chat,"Sell order executed.");
        Print("Sell Order executed.");
        int i=SignalIndex(position.signal_id);
        if(i>=0)
        {
            int d=ArraySize(signals[i].tickets)-1;
            if(d>=0) {ArrayResize(signals[i].tickets, d+2); signals[i].tickets[d]=Trade.RequestPosition();}
        }
        return true;
    }
    bot.SendMessage(bot.user_chat,Trade.ResultRetcodeDescription());

    return false;

}

int orderindex(int id)
{
    int index=-1;
    for(int i=0; i<ArraySize(order_list); i++)
    {
        if(order_list[i].index==id) {index=i;; break;}
    }
    return index;
}

bool TradeCase(order &tradeorder)
{
    //ArrayPrint(order_list);
    if(tradeorder.active) return true;
    static order prev_order;
    string symbol =tradeorder.symbol;

    double Ask = SymbolInfoDouble(symbol, SYMBOL_ASK);

    double Bid = SymbolInfoDouble(symbol, SYMBOL_BID);

    double current_price=tradeorder.type%2>0? Ask:Bid;

    double price=tradeorder.price;

    double lprice=tradeorder.limit_price;

    int i= orderindex(tradeorder.index);

    ordertype type= tradeorder.type;

    bool done;
    //if(prev_order.take_profit==tradeorder.take_profit) return true;
    bool from_top=((type==3) || (type==8) || (type==6) || (type==9));
    bool from_under=((type==4) || (type==5) || (type==7) || (type==10));

    //if(!(((type%3==0 || type==8) && current_price<=price) || ((type>3) && current_price>=price))) return false;

    if(from_top && (current_price>price)) return false;
    else if(from_under && (current_price<price)) return false;

    switch(type)

    {

        case 1: done=m_Buy(tradeorder); order_list[i].active=true;

        break;

        case 2: done=m_Sell(tradeorder); order_list[i].active=true;

        break;

        case 3: done=m_Buy(tradeorder); order_list[i].active=true;

        break;

        case 4: done=m_Sell(tradeorder); order_list[i].active=true;

        break;

        case 5: done=m_Buy(tradeorder); order_list[i].active=true;

        break;

        case 6: done=m_Sell(tradeorder); order_list[i].active=true;

        break;

        case 7: order_list[i].price=lprice; order_list[i].type=3; done=true; order_list[i].expiration=(int)(TimeTradeServer()+((7)*PeriodSeconds(PERIOD_CURRENT)));

        break;

        case 8: order_list[i].price=lprice; order_list[i].type=4; done=true; order_list[i].expiration=(int)(TimeTradeServer()+((7)*PeriodSeconds(PERIOD_CURRENT)));

        break;

        case 9: order_list[i].price=lprice; order_list[i].type=5; done=true; order_list[i].expiration=(int)(TimeTradeServer()+((7)*PeriodSeconds(PERIOD_CURRENT)));

        break;

        case 10: order_list[i].price=lprice; order_list[i].type=6; done=true; order_list[i].expiration=(int)(TimeTradeServer()+((7)*PeriodSeconds(PERIOD_CURRENT)));

        break;

        default: done=false;
    }
    prev_order=tradeorder;
    //ArrayPrint(order_list);
    return done;
}

double m_point(string symbol)
{

    double contract_size = SymbolInfoDouble(symbol,SYMBOL_TRADE_CONTRACT_SIZE);
    double min_vol=SymbolInfoDouble(symbol,SYMBOL_VOLUME_MIN);

    double mul_points= 0.01/(contract_size*min_vol);

    //if((SymbolInfoInteger(symbol, SYMBOL_TRADE_CALC_MODE) == SYMBOL_CALC_MODE_FOREX) || (SymbolInfoInteger(symbol, SYMBOL_TRADE_CALC_MODE) == SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE)) SymbolInfoDouble(symbol,SYMBOL_POINT);
    //if(StringContains(symbol,micro_suffix)) return (mul_points*10);
    return (mul_points);
}

//+------------------------------------------------------------------+

//| Calculates unit cost based on profit calculation mode.           |

//+------------------------------------------------------------------+

double CalculateUnitCost(string symbol)
{
    double UnitCost;

    // CFD.

    if(((CalcMode==SYMBOL_CALC_MODE_CFD) || (CalcMode== SYMBOL_CALC_MODE_CFDINDEX) || (CalcMode == SYMBOL_CALC_MODE_CFDLEVERAGE)))

        UnitCost = SymbolInfoDouble(symbol, SYMBOL_TRADE_TICK_SIZE) * SymbolInfoDouble(symbol, SYMBOL_TRADE_CONTRACT_SIZE);

    // With Forex and futures instruments, tick value already equals 1 unit cost.

    else UnitCost = SymbolInfoDouble(symbol, SYMBOL_TRADE_TICK_VALUE_LOSS);
    return UnitCost;
}



//+-----------------------------------------------------------------------------------+

//| Calculates necessary adjustments for cases when GivenCurrency != AccountCurrency. |

//+-----------------------------------------------------------------------------------+

double CalculateAdjustment(string symbol)

{

    if (ReferencePair == NULL)

    {

        ReferencePair = GetSymbolByCurrencies(ProfitCurrency, AccountCurrency);

        ReferenceSymbolMode = true;

        // Failed.

        if (ReferencePair == NULL)

        {

            // Reversing currencies.

            ReferencePair = GetSymbolByCurrencies(AccountCurrency, ProfitCurrency);

            ReferenceSymbolMode = false;

        }

    }

    if (ReferencePair == NULL)

    {

        Print("Error: Cannot detect proper currency pair for adjustment calculation: ", ProfitCurrency, ", ", AccountCurrency, ".");

        ReferencePair =  symbol;

        return 1;

    }

    MqlTick tick;

    SymbolInfoTick(ReferencePair, tick);

    return GetCurrencyCorrectionCoefficient(tick);
}



//+---------------------------------------------------------------------------+

//| Returns a currency pair with specified base currency and profit currency. |

//+---------------------------------------------------------------------------+

string GetSymbolByCurrencies(string base_currency, string profit_currency)
{
    // Cycle through all symbols.

    for (int s = 0; s < SymbolsTotal(false); s++)

    {

        // Get symbol name by number.

        string symbolname = SymbolName(s, false);

        // Skip non-Forex pairs.

        if ((SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX) && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE)) continue;

        // Get its base currency.

        string b_cur = SymbolInfoString(symbolname, SYMBOL_CURRENCY_BASE);

        if (b_cur == "RUR") b_cur = "RUB";

        // Get its profit currency.

        string p_cur = SymbolInfoString(symbolname, SYMBOL_CURRENCY_PROFIT);

        if (p_cur == "RUR") p_cur = "RUB";



        // If the currency pair matches both currencies, select it in Market Watch and return its name.

        if ((b_cur == base_currency) && (p_cur == profit_currency))

        {

            // Select if necessary.

            if (!(bool)SymbolInfoInteger(symbolname, SYMBOL_SELECT)) SymbolSelect(symbolname, true);

            return symbolname;

        }

    }

    return NULL;

}



//+------------------------------------------------------------------+

//| Get corLTrection coefficient based on currency, trade direction,   |

//| and current prices.                                              |

//+------------------------------------------------------------------+

double GetCurrencyCorrectionCoefficient(MqlTick &tick)

{

    if ((tick.ask == 0) || (tick.bid == 0)) return -1; // Data is not yet ready.

    // Reverse quote.

    if (ReferenceSymbolMode)

    {

        // Using Buy price for reverse quote.

        return tick.ask;

    }

    // Direct quote.

    else

    {

        // Using Sell price for direct quote.

        return (1 / tick.bid);

    }

}


//+------------------------------------------------------------------+

//| Calculate position size depending on money management parameters.|

//+------------------------------------------------------------------+

double LotsOptimized(string symbol, ENUM_ORDER_TYPE Dir, double pending_entry = 0, double SL=0.0)

{
    double Size, RiskMoney, PositionSize = 0;
    if(SL==0) return SymbolInfoDouble( symbol, SYMBOL_VOLUME_MIN);
    
        double LotStep = SymbolInfoDouble( symbol, SYMBOL_VOLUME_STEP);
    
        int LotStep_digits = CountDecimalPlaces(LotStep);
    
    if(!MM) PositionSize=Lots;
    else
    {
        if (AccountInfoString(ACCOUNT_CURRENCY) == "") return 0;
    
        if (FixedBalance > 0)
    
        {
    
            Size = FixedBalance;
    
        }
    
        else if (UseEquityInsteadOfBalance)
    
        {
    
            Size = AccountInfoDouble(ACCOUNT_EQUITY);
    
        }
    
        else
    
        {
    
            Size = AccountInfoDouble(ACCOUNT_BALANCE);
    
        }
    
        if (!UseMoneyInsteadOfPercentage) RiskMoney = Size * Risk / 100;
    
        else RiskMoney = MoneyRisk;
    
        double UnitCost = CalculateUnitCost(symbol);
    
        // If profit currency is different from account currency and Symbol is not a Forex pair or futures (CFD, and so on).
    
        if ((ProfitCurrency != AccountCurrency) && (CalcMode != SYMBOL_CALC_MODE_FOREX) && (CalcMode != SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE) && (CalcMode != SYMBOL_CALC_MODE_FUTURES) && (CalcMode != SYMBOL_CALC_MODE_EXCH_FUTURES) && (CalcMode != SYMBOL_CALC_MODE_EXCH_FUTURES_FORTS))
    
        {
    
            double CCC = CalculateAdjustment(symbol); // Valid only for loss calculation.
    
            // Adjust the unit cost.
    
            UnitCost *= CCC;
    
        }
    
        // If account currency == pair's base currency, adjust UnitCost to future rate (SL). Works only for Forex pairs.
    
        if ((AccountCurrency == BaseCurrency) && ((CalcMode == SYMBOL_CALC_MODE_FOREX) || (CalcMode == SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE)))
    
        {
    
            double current_rate = 1, future_rate = 1;
    
            if (Dir == ORDER_TYPE_BUY)
    
            {
    
                if (pending_entry == 0) current_rate = SymbolInfoDouble( symbol, SYMBOL_ASK);
    
                else current_rate = pending_entry;
    
                future_rate = current_rate - SL;
    
            }
    
            else if (Dir == ORDER_TYPE_SELL)
    
            {
    
                if (pending_entry == 0) current_rate = SymbolInfoDouble( symbol, SYMBOL_BID);
    
                else current_rate = pending_entry;
    
                future_rate = current_rate + SL;
    
            }
    
            if (future_rate == 0) future_rate = SymbolInfoDouble( symbol, SYMBOL_POINT); // Zero divide prevention.
    
            UnitCost *= (current_rate / future_rate);
    
        }
    
    
    
        double TickSize = SymbolInfoDouble( symbol, SYMBOL_TRADE_TICK_SIZE);
    
        double LotStep = SymbolInfoDouble( symbol, SYMBOL_VOLUME_STEP);
    
        int LotStep_digits = CountDecimalPlaces(LotStep);
    
        if ((SL != 0) && (UnitCost != 0) && (TickSize != 0)) PositionSize = NormalizeDouble(RiskMoney / (SL * UnitCost / TickSize), LotStep_digits);
    
    
    
        if (PositionSize < SymbolInfoDouble( symbol, SYMBOL_VOLUME_MIN))
    
        {
            if(no_money_no_trade) return 0;
            Print("Calculated position size (" + DoubleToString(PositionSize, 2) + ") is less than minimum position size (" + DoubleToString(SymbolInfoDouble( symbol, SYMBOL_VOLUME_MIN), 2) + "). Setting position size to minimum.");
    
            PositionSize = SymbolInfoDouble( symbol, SYMBOL_VOLUME_MIN);
    
        }
    
        else if ((PositionSize < UserMinLot) && (UserMinLot>=SymbolInfoDouble( symbol, SYMBOL_VOLUME_MIN)))
    
        {
    
            Print("Calculated position size (" + DoubleToString(PositionSize, 2) + ") is less than the specified minimum position size (" + DoubleToString(UserMinLot, 2) + "). Setting position size to the specified minimum.");
    
            PositionSize = UserMinLot;
    
        }
    
        else if ((PositionSize > UserMaxLot) && (UserMaxLot<=SymbolInfoDouble( symbol, SYMBOL_VOLUME_MAX)))
    
        {
    
            Print("Calculated position size (" + DoubleToString(PositionSize, 2) + ") is greater than the specified maximum position size (" + DoubleToString(UserMaxLot, 2) + "). Setting position size to the specified maximum.");
    
            PositionSize = UserMaxLot;
    
        }
    
        else if (PositionSize > SymbolInfoDouble( symbol, SYMBOL_VOLUME_MAX))
    
        {
    
            Print("Calculated position size (" + DoubleToString(PositionSize, 2) + ") is greater than maximum position size (" + DoubleToString(SymbolInfoDouble( symbol, SYMBOL_VOLUME_MAX), 2) + "). Setting position size to maximum.");
    
            PositionSize = SymbolInfoDouble( symbol, SYMBOL_VOLUME_MAX);
    
        }
    
    
    
        double steps = PositionSize / LotStep;
    
        if (MathFloor(steps) < steps)
    
        {
    
            Print("Calculated position size (" + DoubleToString(PositionSize, 2) + ") uses uneven step size. Allowed step size = " + DoubleToString(SymbolInfoDouble( symbol, SYMBOL_VOLUME_STEP), 2) + ". Setting position size to " + DoubleToString(MathFloor(steps) * LotStep, 2) + ".");
    
            PositionSize = MathFloor(steps) * LotStep;
    
        }
    }
    //PositionSize=NormalizeDouble(PositionSize*0.5, LotStep_digits);
    if (PositionSize < SymbolInfoDouble( symbol, SYMBOL_VOLUME_MIN))
    {
        if(no_money_no_trade) return 0;
        Print("Specified position size (" + DoubleToString(PositionSize, 2) + ") is less than minimum position size (" + DoubleToString(SymbolInfoDouble( symbol, SYMBOL_VOLUME_MIN), 2) + "). Setting position size to minimum.");
        PositionSize = SymbolInfoDouble( symbol, SYMBOL_VOLUME_MIN);
    }
    else if ((PositionSize < UserMinLot) && (UserMinLot>=SymbolInfoDouble( symbol, SYMBOL_VOLUME_MIN)))
    {
        Print("Specified position size (" + DoubleToString(PositionSize, 2) + ") is less than the specified minimum position size (" + DoubleToString(UserMinLot, 2) + "). Setting position size to the specified minimum.");
        PositionSize = UserMinLot;
    }
    return PositionSize;
}

//+------------------------------------------------------------------+

//| Counts decimal places.                                           |

//+------------------------------------------------------------------+

int CountDecimalPlaces(double number)
{
    // 100 as maximum length of number.
    for (int i = 0; i < 100; i++)
    {
        double pwr = MathPow(10, i);
        if (MathRound(number * pwr) / pwr == number) return i;
    }

    return -1;
}



bool IsNaN(string &text)
{
    double num=(double)(text);
    if((num>0) && (num<999999)) return true;
    return false;
}

//----------------------

string ReformText(string txt)
{
    string msg = StringToLowerCase(txt);
    //Lists of keywords & mappings
    string buy[], sell[], boom1000[], boom500[], boom300[], crash1000[], crash500[];
    string crash300[], vix100[], vix100s[], vix50[], vix50s[], vix25[], vix25s[];
    string vix10[], vix10s[], vix75[], vix75s[], stepindex[], close_partials[], action_close[];
    string action_be[], move_tp[], re_entry[];
    //string extras[];
    
    //StringSplitTrim(StringToLowerCase(keyword),",", extras);
    StringSplitTrim(StringToLowerCase("book some profits,take partial,close half"),",", close_partials);
    StringSplitTrim(StringToLowerCase("cutloss,cut loss,close all,delete"),",", action_close);
    StringSplitTrim(StringToLowerCase("move sl,move stoploss,edit stoploss"),",", action_be);
    StringSplitTrim(StringToLowerCase("move tp,move takeprofit,move tp to,move takeprofit to,close at entry"),",", move_tp);
    StringSplitTrim(StringToLowerCase("re-enter,re entry,re enter,buy again,sell again,resell,rebuy"),",", re_entry);
    StringSplitTrim(StringToLowerCase("long,buy"),",", buy);
    StringSplitTrim(StringToLowerCase("short,sell"),",", sell);
    StringSplitTrim(StringToLowerCase("Boom 1000, boom1000"),",", boom1000);
    StringSplitTrim(StringToLowerCase("boom 500,boom500"),",", boom500);
    StringSplitTrim(StringToLowerCase("boom 300,boom300"),",", boom300);
    StringSplitTrim(StringToLowerCase("crash 1000"),",", crash1000);
    StringSplitTrim(StringToLowerCase("crash 500"),",", crash500);
    StringSplitTrim(StringToLowerCase("crash 300"),",", crash300);
    StringSplitTrim(StringToLowerCase("step index"),",", stepindex);
    StringSplitTrim(StringToLowerCase("vix 100, volatility 100, volatility100"),",", vix100);
    StringSplitTrim(StringToLowerCase("vix 100s, vix100 s, vix 100 s, volatility 100s, volatility 100 1s, volatility100s, volatility 100 s"),",", vix100s);
    StringSplitTrim(StringToLowerCase("vix 75, volatility 75, volatility75"),",", vix75);
    StringSplitTrim(StringToLowerCase("vix 75s, vix 75 s, vix75 s, volatility 75s, volatility 75 1s, volatility75s, volatility 75 s"),",", vix75s);
    StringSplitTrim(StringToLowerCase("vix 50,volatility 50, volatility50"),",", vix50);
    StringSplitTrim(StringToLowerCase("vix 50s,volatility 50s, vix50 s, vix 50 s,volatility 50 1s, volatility50s, volatility 50 s"),",", vix50s);
    StringSplitTrim(StringToLowerCase("vix 25,volatility 25, volatility25"),",", vix25);
    StringSplitTrim(StringToLowerCase("vix 25s, vix25 s, vix 25 s, volatility 25s, volatility 25 1s, volatility25s, volatility 25 s"),",", vix25s);
    StringSplitTrim(StringToLowerCase("vix 10,volatility 10, volatility10"),",", vix10);
    StringSplitTrim(StringToLowerCase("vix 10s, vix10 s, vix 10 s, volatility 10s, volatility 10 1s, volatility10s, volatility 10 s"),",", vix10s);
    
    //Remove Special characters 
    msg = StringReplace2(msg," @ "," at ");
    msg = StringReplace2(msg,"@"," at ");
    msg = StringReplace2(msg,":"," ");
    msg = StringReplace2(msg,"("," ");
    msg = StringReplace2(msg,")"," ");
    msg = StringReplace2(msg,";"," ");
    msg = StringReplace2(msg,"/"," ");
    msg = StringReplace2(msg,"-"," ");
    msg = StringReplace2(msg,"_"," ");
    msg = StringReplace2(msg,"..."," ");
    msg = StringReplace2(msg,".. "," ");
    msg = StringReplace2(msg,". "," ");
    
    //Replace key-texts
    msg = StringReplace2(msg,"limit","entry");
    msg = StringReplace2(msg,"xauusd","gold");
    msg = StringReplace2(msg,"target","tp");
    msg = StringReplace2(msg,"sl ","stoploss ");
    msg = StringReplace2(msg,"stop loss","stoploss ");
    msg = StringReplace2(msg,"stoploss","stoploss ");
    msg = StringReplace2(msg,"buy stop","buystop ");
    msg = StringReplace2(msg,"sell stop","sellstop ");
    //msg = StringReplace2(msg,"buystop","buystop ");
    //msg = StringReplace2(msg,"sellstop","sellstop ");
    msg = StringReplace2(msg," now "," entry ");
    msg = StringReplace2(msg,"buy at","buy entry ");
    msg = StringReplace2(msg,"sell at","sell entry ");
    msg = StringReplace2(msg,"take profit","takeprofit ");
    msg = StringReplace2(msg,"takeprofit","tp ");
    msg = ReplaceFromList(msg, close_partials,"close half");
    msg = ReplaceFromList(msg, action_close,"close all");
    msg = ReplaceFromList(msg, action_be,"breakeven");
    msg = ReplaceFromList(msg, move_tp,"move tp");
    msg = ReplaceFromList(msg, re_entry,"re-enter");
    
    msg = StringReplace2(msg,"\n"," ");
    
    msg = StringReplace2(msg,"tp 1 ","tp1 ");
    msg = StringReplace2(msg,"tp 2 ","tp2 ");
    msg = StringReplace2(msg,"tp 3 ","tp3 ");
    msg = StringReplace2(msg,"tp 4 ","tp4 ");
    msg = StringReplace2(msg,"tp 5 ","tp5 ");
    if(StringContains(msg,"tp2 ")) msg = StringReplace2(msg,"tp ","tp1 ");
    while(StringContains(msg,"  ")) msg = StringReplace2(msg,"  "," ");
    msg = StringReplace2(msg,"open","0");
    msg = ReplaceFromList(msg, buy,"buy ");
    msg = ReplaceFromList(msg, sell,"sell ");
    if((StringContains(msg," stop")) && !(StringContains(msg,"loss "))) msg = StringReplace2(msg,"stop","stoploss ");
    if((StringContains(msg,"sell at")) || (StringContains(msg,"buy at"))) msg = StringReplace2(msg," at"," entry ");
    msg = StringReplace2(msg," at "," entry ");
    msg = StringReplace2(msg,"ep "," entry ");
    msg = StringReplace2(msg," price "," ");
    msg = StringReplace2(msg,"entry entry"," entry ");
    //if((StringContains(msg,"sell")) && (StringContains(msg,"limit"))) msg = StringReplace2(msg,"sell","sell&limit ");
    //if((StringContains(msg,"buy")) && (StringContains(msg,"limit"))) msg = StringReplace2(msg,"buy","buy&limit ");
    
    //For Trading instruments
    msg = ReplaceFromList(msg, boom1000,"Boom1000");
    msg = ReplaceFromList(msg, boom500,"Boom500");
    msg = ReplaceFromList(msg, boom300,"Boom300");
    msg = ReplaceFromList(msg, crash1000,"Crash1000");
    msg = ReplaceFromList(msg, crash500,"Crash500");
    msg = ReplaceFromList(msg, crash300,"Crash300");
    msg = ReplaceFromList(msg, stepindex,"StepIndex");
    msg = ReplaceFromList(msg, vix100,"Vix100");
    msg = ReplaceFromList(msg, vix100s,"Vix100s");
    msg = ReplaceFromList(msg, vix75,"Vix75");
    msg = ReplaceFromList(msg, vix75s,"Vix75s");
    msg = ReplaceFromList(msg, vix50,"Vix50");
    msg = ReplaceFromList(msg, vix50s,"Vix50s");
    msg = ReplaceFromList(msg, vix25,"Vix25");
    msg = ReplaceFromList(msg, vix25s,"Vix25s");
    msg = ReplaceFromList(msg, vix10,"Vix10");
    msg = ReplaceFromList(msg, vix10s,"Vix10s");
    int m=ArraySize(pair_key);
    while(--m>=0) 
    {
        string attributes[];
        StringSplitTrim(pair_attribute[m],"-",attributes);
        ReplaceFromList(msg, attributes,pair_key[m]);
    }
    
//--
    
    msg = StringReplace2(msg,"  "," ");
    //bot.SendMessage(bot.user_chat,msg);
    return msg;
}

string ReplaceFromList(string msg, string &list[], string replacement)
{
    int items = ArraySize(list);
    for(int i=0; i<items; i++)
    {
        msg = StringReplace2(msg,StringTrim(list[i]),replacement);
    }
    return msg;
}

double FindV(string msg, string keyword, bool range=false)
{
    double value=0, value2=0;
    string imsg=msg;
    if(StringContains(imsg, keyword))
    {
        int i =0;
        string word[];
        int num = StringSplitTrim(imsg, " ", word);
        while(StringContains(imsg, keyword) && (i<10))
        {
            int keyword_index = StringIndexOf(imsg, keyword);
            int idx = StringCountMatches(StringSubstr(imsg,0, keyword_index), " ") + 1;
            if(idx<ArraySize(word)) 
            {
                if(!StringIsNumeric(word[idx]) && ((word[idx]!="stoploss") || (word[idx]!="tp"))) {imsg=StringRemove(imsg, word[idx]); Print(imsg," ...and na me dey do am o"); StringSplitTrim(imsg, " ", word); i = i+1; continue;}
            }
            if(range && value!=0) value2=StringToDouble(word[idx]);
            else if(idx<ArraySize(word)) value = StringToDouble(word[idx]);
            if(idx>=ArraySize(word)) break;
            //Clear
            string remove= StringSubstr(imsg,0,StringIndexOf(imsg,keyword,0)+StringLen(keyword));
            //Print(remove,"...removed");
            //msg=StringRemove(msg, remove);
            StringSplitTrim(imsg, " ", word);
            //--
            i=i+1;
            //i = i+1;
        }
    }
    imsg=msg;
    if(range) return value2;
    return value;
}

int FindW(string msg, bool type = false)
{
    string text = StringToLowerCase(msg);
    
    /** For trade type: 1==>buy, 2==>sell, 3==>buylimit, 4==>selllimit, 5==> buy&limit, 6==>sell&limit, 7==>buystop, 8==>sellstop, 9==>buy&stop, 10==>sell&stop,
    For symbols: 1==>Vix100, 2==>Vix100s, 3==>Vix75, 4==>Vix75s, 5==>Vix50, 6==>Vix50s, 7==>Vix25, 8==>Vix25s, 9==>Vix10, 10==>Vix10s, 
    11==>Boom1000, 12==>Boom500, 13==>Boom300
    14==>Crash1000, 15==>Crash500, 16==>Crash300
    17==>StepIndex,
    **/
    if(!type)
    {
        if(StringContains(text, "vix100 ")) return 1; 
        if(StringContains(text, "vix100s ")) return 2;
        if(StringContains(text, "vix75 ")) return 3;
        if(StringContains(text, "vix75s ")) return 4;
        if(StringContains(text, "vix50 ")) return 5;
        if(StringContains(text, "vix50s ")) return 6;
        if(StringContains(text, "vix25 ")) return 7;
        if(StringContains(text, "vix25s ")) return 8;
        if(StringContains(text, "vix10 ")) return 9;
        if(StringContains(text, "vix10s ")) return 10;
        if(StringContains(text, "boom1000 ")) return 11;
        if(StringContains(text, "boom500 ")) return 12;
        if(StringContains(text, "boom300 ")) return 13;
        if(StringContains(text, "crash1000 ")) return 14;
        if(StringContains(text, "crash500 ")) return 15;
        if(StringContains(text, "crash300 ")) return 16;
        if(StringContains(text, "stepindex ")) return 17;
    }
    else
    {
        if(StringContains(text, "entry ")) return 0;
        if(StringContains(text, "buy ")) return 18;
        if(StringContains(text, "sell ")) return 21;
    }
    return 0;
}

double RrToPrice(string symbol, double open, double secure_rr, double sl_pip, bool type)
{
    double contract_size = SymbolInfoDouble(symbol,SYMBOL_TRADE_CONTRACT_SIZE);
    double points = (sl_pip*secure_rr)/(contract_size*0.1);
    if(type) return open + points;
    else return open - points;
}

//Add recognition of forex pairs and stocks
string forexpairs(string text)
{
    string textp[];
    StringSplitTrim(StringToUpperCase(text)," ",textp);
    //ArrayPrint(textp);
    for(int i=0; i <ArraySize(textp); i++)
    {
        //Sorting for Forex pairs. 
        string sym=textp[i];
        //Print(sym);
        if (StringLen(textp[i])>3)
        { 
            for (int s = 0; s < SymbolsTotal(true); s++)
            { 
                //identify suffixes for forex micro pairs
                
                //if(SymbolInfoDouble("EURUSD",SYMBOL_TRADE_CONTRACT_SIZE)==1000;
                string symbolname =SymbolName(s, true);
                //Print(symbolname);
                 // Skip non-Forex pairs.
                //if ((SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX) && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE)) continue;
                if((StringContains(symbolname,sym)) || (sym=="GOLD" && StringContains(symbolname,"XAUUSD")))
                { 
                    //if (!(bool)SymbolInfoInteger(symbolname, SYMBOL_SELECT)) SymbolSelect(symbolname, true);
                    if(SymbolInfoInteger(symbolname,SYMBOL_TRADE_MODE)==SYMBOL_TRADE_MODE_DISABLED) continue;
                    return symbolname; 
                } 
            }
            for (int s = 0; s < SymbolsTotal(false); s++)
            { 
                //identify suffixes for forex micro pairs
                
                //if(SymbolInfoDouble("EURUSD",SYMBOL_TRADE_CONTRACT_SIZE)==1000;
                string symbolname =SymbolName(s, false);
                string s_name= StringSubstr(symbolname,0,6);
                //Print(s_name);
                 // Skip non-Forex pairs.
                //if ((SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX) && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE)) continue;
                if((StringContains(symbolname,sym)) || (sym=="GOLD" && StringContains(symbolname,"XAUUSD")))
                { 
                    //--
                    if(SymbolInfoInteger(symbolname,SYMBOL_TRADE_MODE)==SYMBOL_TRADE_MODE_DISABLED) continue;
                    if (!(bool)SymbolInfoInteger(symbolname, SYMBOL_SELECT)) SymbolSelect(symbolname, true);
                    return symbolname; 
                } 
            }
        }
        }
        //Guess the symbol by price. 
        double pricem=FindV(text, "stoploss");
        if(pricem<=0) return "";
        string suffix ; 
        double lowest=999999.9;
        double price[];
        string symbol[];
        //if(IsMicro(suffix ) && StringContains(suffix, symbolname))
        
        for (int s = 0; s < SymbolsTotal(false); s++)
        { 
                //if(SymbolInfoDouble("EURUSD",SYMBOL_TRADE_CONTRACT_SIZE)==1000;
            string symbolname =SymbolName(s, false);
            if ((SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX) 
            && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_CFD)
            && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_CFDLEVERAGE) 
            && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FUTURES) 
            && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_CFDINDEX)  
            && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE)) continue;
            if(SymbolInfoInteger(symbolname,SYMBOL_TRADE_MODE)==SYMBOL_TRADE_MODE_DISABLED) continue;
            if (!(bool)SymbolInfoInteger(symbolname, SYMBOL_SELECT)) SymbolSelect(symbolname, true);
        }
        for (int s = 0; s < SymbolsTotal(true); s++)
        {  
            string symbolname = SymbolName(s, true);
            if ((SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX) && (SymbolInfoInteger(symbolname, SYMBOL_TRADE_CALC_MODE) != SYMBOL_CALC_MODE_FOREX_NO_LEVERAGE)) continue;
        
            double cmp = SymbolInfoDouble(symbolname, SYMBOL_ASK);
            double vol=micro?10:0.1;
            ArrayResize(price,ArraySize(price)+1);
            price[ArraySize(price)-1]=(CostPips(symbolname, cmp, pricem, vol));
            ArrayResize(symbol,ArraySize(symbol)+1);
            symbol[ArraySize(symbol)-1]=symbolname;
        }
        //Select the closest match.
        //ArrayPrint(price);
        int low_index=-1;
        for(int i=0; i<ArraySize(price); i++)
        {
            if(price[i]<lowest) {lowest = price[i]; low_index=i;}
            
        }
        if((low_index<0) || (low_index>=ArraySize(symbol))) return "";
        if (!(bool)SymbolInfoInteger(symbol[low_index], SYMBOL_SELECT)) SymbolSelect(symbol[low_index], true);
        return symbol[low_index];
    
    return "";
}

string Instrument(int code, string msg)
{
    string pair=forexpairs(msg);
    if(pair!="") return pair;
    string instrument [17];
    instrument[0] = "Volatility 100 Index";
    instrument[1] = "Volatility 100 (1s) Index";
    instrument[2] = "Volatility 75 Index";
    instrument[3] = "Volatility 75 (1s) Index";
    instrument[4] = "Volatility 50 Index";
    instrument[5] = "Volatility 50 (1s) Index";
    instrument[6] = "Volatility 25 Index";
    instrument[7] = "Volatility 25 (1s) Index";
    instrument[8] = "Volatility 10 Index";
    instrument[9] = "Volatility 10 (1s) Index";
    instrument[10] = "Boom 1000 Index";
    instrument[11] = "Boom 500 Index";
    instrument[12] = "Boom 300 Index";
    instrument[13] = "Crash 1000 Index";
    instrument[14] = "Crash 500 Index";
    instrument[15] = "Crash 300 Index";
    instrument[16] = "Step Index";
    
    if(code<=0)
    {
        for(int i=0; i<17; i++)
        {
            double pips = CostPips(instrument[i], fmax(FindV(msg, "limit"),FindV(msg, "stoploss")), SymbolInfoDouble(instrument[i],SYMBOL_BID), SymbolInfoDouble(instrument[i], SYMBOL_VOLUME_MIN));
            if((pips<=100) && pips>5)return instrument[i];
        }
    }
    else if(code<=17) return instrument[code-1];
    return "";
}

bool GetStops(string text, double cmp, double &sl, double &tp, double &tps[], double &entry)
{
    sl=entry=tp=0;
    string msg = text;
    sl = FindV(msg, "stoploss");
    msg=text;
    //double tps[];
    //ArrayResize(tps, 5);
    ArrayInitialize(tps, 0.0);
    entry = FindV(msg, "entry");
    /**
    double limit1 = FindV(msg, "entry");
    double limit2 = FindV(msg, "entry", true);
    msg=text;
    //double limit;
    if((limit1!=limit2) && (limit1!=0))
    {
        if(limit2!=0) entry = (limit1+limit2)/2;
        else entry = (limit1+cmp)/2;
    }**/
    if(StringContains(msg, "tp "))
    {
        int i =0;
        string word=msg;
        while(StringContains(word, "tp "))
        {
            //Print(word);
            ArrayResize(tps,i+1);
            tps[i] = FindV(word,"tp");
            //Clear
            int inde=StringIndexOf(word,"tp ",0);
            string remove=StringSubstr(word,inde,inde+2),remove2=(string)(tps[i]);
            word=StringRemove(word,remove);
            if(tps[i]>0) word=StringRemove(word,remove2);
            Print(remove,remove2,"...taken out in:",word);
            //--
            i = i+1;
            //Print(word);
        }
    }
    else if(StringContains(msg, "tp"))
    {
        ArrayResize(tps,5);
        tps[0]= FindV(msg, "tp1");
        //tps[0]=tps[0]==0? FindV(msg, "tp"):tps[0];
        tps[1]= FindV(msg, "tp2");
        tps[2]= FindV(msg, "tp3");
        tps[3]= FindV(msg, "tp4");
        tps[4]= FindV(msg, "tp5");
        if(tps[0]>0)
        {
            if(tps[1]>0)
            {
                if(tps[2]>0)
                {
                    if(tps[3]>0)
                    {
                        if(tps[4]==0) ArrayResize(tps,4);
                    }
                    else ArrayResize(tps,3);
                }
                else ArrayResize(tps,2);
            }
            else ArrayResize(tps,1);
        }
    }
     
    tp=0;
    //ArrayPrint(tps);
    //if(sl==0) {tp=tps[0]; return false;}
    int z=ArraySize(tps)-1;
    
    if(z>=0) tp=tps[z];
    return sl!=0;
}


signal_action GetAction(string msg)
{
    string text=msg;
    double tp=FindV(text,"tp");
    double sl=FindV(text, "stoploss");
    string symbol= Instrument(FindW(text,true),text);
    if(symbol=="")
    {
        bot.SendMessage(bot.user_chat,"On signal_action! check text below.");
        bot.SendMessage(bot.user_chat,text);
        if(StringContains(text,"stoploss")) sl=GetDoubleValue(text);
        if(StringContains(text,"tp ")) tp=GetDoubleValue(text);
        if(tp!=0) return 3;
        if(sl!=0) return 1;
    }
    if(StringContains(msg,"breakeven") && StringContains(msg,"close half")) return 5;
    else if(StringContains(msg,"close half")) return 4;
    else if(StringContains(msg,"move tp")) return 3;
    else if(StringContains(msg,"close")) return 2;
    else if(StringContains(msg,"breakeven")) return 1;
    else return 0;
}

long SignalId(long msg_id)
{
    //Search signal array.
    long id=-1;
    int no_of_signals=ArraySize(signals);
    for(int i=no_of_signals-1; i>=0; i--)
    {
        if(signals[i].message_id==msg_id) {id=signals[i].signal_id; break;}
        else
        {
            //Search message ids
            int no_of_ids=ArraySize(signals[i].message_ids);
            for(int j=0; j<no_of_ids; j++)
            {
                long m_id=signals[i].message_ids[j];
                if(m_id==msg_id) {id=signals[i].signal_id; break;}
            }
        }
    }
    return id;
}

long SignalId(string msg_text)
{
    long id=-1;
    string text=ReformText(msg_text);
    id=(long)(FindV(text,"id"));
    return id;
}

int SignalIndex(long signal_id)
{
    int index=-1;
    int no_of_signals=ArraySize(signals);
    for(int i=no_of_signals-1; i>=0; i--)
    {
        if(signals[i].signal_id==signal_id) {index=i; break;}
    }
    return index;
   }
   
bool ValidityCheck(string &text, CCustomMessage &message)
{
    Print("Valid point!");
    text=ReformText(text);
    signal new_item;
    int i=ArraySize(signals);
    new_item.symbol=Instrument(FindW(text,true),text);
    new_item.signal_id=i+100000;
   //int l=ArraySize(signals)-1;
    new_item._comment=(string)(StringToUpperCase(message.chat_username)); 
    new_item.message_id=message.message_id;
    new_item.signal_time=TimeLocal();
   //int a=ArraySize(sources);
    return bot.ProcessAction(new_item,0,text);
    
}

void GetOrders(string &msg[])
{
    
    //-- Pending Orders
    for(int i=0; i<ArraySize(order_list); i++)
    {
        order pending=order_list[i];
        if(pending.active==true) continue;
        ArrayResize(msg,ArraySize(msg)+1);
        int s_id=SignalIndex(pending.signal_id);
        string _comment=pending.symbol +" "+ EnumToString(pending.type) + "\n";
        double bid=SymbolInfoDouble(pending.symbol,SYMBOL_BID);
        _comment+= "Entry: " +(string)(pending.price) + "\n";
        _comment+= "Take Profit: " + (string)(pending.take_profit) + "\n";
        _comment+= "StopLoss: " + (string)(pending.stoploss) + "\n";
        _comment+= "Current Price: " +(string)(bid) + "\n";
        _comment+= "Setup Time: "+ TimeToString(pending.setup_time) +"\n";
        _comment+= "Signal from ";
        if(s_id>-1) _comment+= signals[s_id]._comment + "\n";
        else _comment+="Unknown Source." + "\n";
        _comment+= "Signal Id: " +(string)(pending.signal_id);
        msg[ArraySize(msg)-1]=_comment;
    }
    //--
}

bool GetPositions(string &msgs[])
{
    CPositionInfo position;
    int total=PositionsTotal();
    for(int i=total-1; i>=0; i--)
    {
        position.SelectByIndex(i);
        if(position.Magic()!=Magic) continue;
        string comment_=position.Comment();
        long id=PosSignalId(comment_);
        if(!id) continue;
        int s_id=SignalIndex(id);
        
        ArrayResize(msgs,ArraySize(msgs)+1);
        string dir=position.PositionType()==POSITION_TYPE_BUY? " BUY":" SELL";
        string _comment=position.Symbol() + dir + "\n";
        _comment+= "Entry: " +(string)(position.PriceOpen()) + "\n";
        _comment+= "Take Profit: " + (string)(position.TakeProfit()) + "\n";
        _comment+= "StopLoss: " + (string)(position.StopLoss()) + "\n";
        _comment+= "Profit: " +(string)(position.Profit()) + "\n";
        _comment+= "Current Price: " +(string)(position.PriceCurrent()) + "\n";
        _comment+= "Position Size: " +(string)(position.Volume())+" lots." + "\n";
        _comment+= "Signal from ";
        if(s_id>-1) _comment+= signals[s_id]._comment + "\n";
        else _comment+="Unknown Source." + "\n";
        _comment+= "Signal Id: " +(string)(id);
        msgs[ArraySize(msgs)-1]=_comment;
    }
    return ArraySize(msgs)>0;
}


double CostPips(string symbol, double entry, double exit, double volume)
{
    double contract_size = SymbolInfoDouble(symbol,SYMBOL_TRADE_CONTRACT_SIZE);
    return (fabs(entry-exit) * contract_size * volume);
}


void BreakEven()
{
    int total = PositionsTotal();
    for (int cnt = total - 1; cnt >= 0; cnt--)
    {
        string symbol = PositionGetSymbol(cnt);
        //if (!PositionSelect(symbol)) continue;
        ulong ticket = PositionGetInteger(POSITION_TICKET);
        string comment = PositionGetString(POSITION_COMMENT);
        //if (PositionGetInteger(POSITION_MAGIC) != ulong(comment)) continue;
        long signal_id=PosSignalId(comment);
        if(!signal_id) continue;
        int index=SignalIndex(signal_id);
        if(index==-1) continue;
        
        double cmp = SymbolInfoDouble(symbol, SYMBOL_BID);
        double open =PositionGetDouble(POSITION_PRICE_OPEN);
        double sl = PositionGetDouble(POSITION_SL);
        bool type = PositionGetInteger(POSITION_TYPE) == POSITION_TYPE_BUY;
        
        double be_price=(2*open)-sl;    //signals[index].tp;
        if((ArraySize(signals[index].tps)>be_tp) && (ArraySize(signals[index].tps)>0)) be_price=signals[index].tps[(int)(be_tp)];
        else if(signals[index].tp>0) be_price=signals[index].tp;
        double sl_pip = CostPips(symbol, open, sl, 0.1);
        double cmp_pips = CostPips(symbol, open, cmp, 0.1);
        double rr = cmp_pips/sl_pip;
        string dir=type?" BUY":" SELL";
        string return_text="BreakEven set on " + symbol + dir;
        //BE_rr is the threshold rr for setting breakeven
        //Inp_Srr is the indicated rr in profit to secure 
        
        double secure_rr = 0;
        if((break_at_rr &&(rr>be_rr))
        || ((type && cmp>be_price) || ((!type) && cmp<be_price)))
        {
            if(secure_rr> (0.66* be_rr)) secure_rr = 0.33 * be_rr;
            double new_sl;
            if(secure_rr<=0) new_sl =open;
            else new_sl = RrToPrice(symbol, open, secure_rr, sl_pip, type);
            if((type && PositionGetDouble(POSITION_SL)>=new_sl) || ((!type) && PositionGetDouble(POSITION_SL)<=new_sl))
            {
                Print(symbol, ": Moving SL to breakeven: ", new_sl, ".");
                if(!mult_pos) bot.ProcessAction(signals[index],4,comment);
                for (int i = 0; i < 10; i++)
                {
                    bool result = Trade.PositionModify(ticket, new_sl, PositionGetDouble(POSITION_TP));
                    if (result) {bot.SendMessage(bot.user_chat,return_text); break;}
                }
            }
        }
    }
}


long last_signal(string &username)
{
    int s=ArraySize(signals)-1;
    for(int i=s; i>=0; i--)
    {
        if(signals[i]._comment!=StringToUpperCase(username)) continue;
        return signals[i].signal_id;
    }
    return 0;
}

long PosSignalId(string comment)
{
    long signal_id;
    if(StringContains(comment,"from"))
    {
        string pieces[];
        ulong ticket=0;
        StringSplitTrim(comment," ",pieces);
        int n=ArraySize(pieces);
        while(--n>=0) {if(StringIsNumeric(pieces[n])) {ticket=(ulong)(pieces[n]); break;}}
        string comment2= HistoryDealGetString(ticket,DEAL_COMMENT);
        bot.SendMessage(bot.user_chat, comment2);
        if(ticket==0) bot.SendMessage(bot.user_chat, comment);
        signal_id=PosSignalId(comment2);
    }
    signal_id=(long)(StringSubstr(comment,0,6));
    if(signal_id<100000 || signal_id>999999) return -1;
    return signal_id;
}

bool SendSignalStats(int duration_days=0)
{
    CDealInfo deal;
    int total=0;
    int losing=0;
    int winning=0;
    //--
    string msg="--- Summary of Trades --- \n";
    //--
    datetime to=TimeCurrent();
    datetime from=to-(PeriodSeconds(PERIOD_D1)*duration_days);
    if(duration_days==0) from=iTime(_Symbol,PERIOD_D1,0);
    int orders;
    if(HistorySelect(from, to)) orders=HistoryDealsTotal();
    else return false;
    for(int i=0; i<orders; i++)
    {
        deal.SelectByIndex(i);
        if((deal.Entry()!=DEAL_ENTRY_OUT) || ((deal.DealType()!=DEAL_TYPE_BUY) && (deal.DealType()!=DEAL_TYPE_SELL))) continue;
        long signal_id=PosSignalId(deal.Comment());
        int index=SignalIndex(signal_id);
        if(!index) continue;
        total+=1;
        string username=signals[index]._comment;
        //--
        string symbol=deal.Symbol();
        double profit=deal.Profit();
        if(profit>=0) winning+=1;
        else losing+=1;
        string type=deal.DealType()==DEAL_TYPE_BUY? " Sell ":" Buy ";
        msg+= symbol + type +"profit: "+(string)(profit) +"\n";
    }
    if(total==0) return false;
    msg+="Total trades: "+(string)(total) +"\n";
    msg+="Winning trades: "+(string)(winning) +"\n";
    msg+="Win rate: "+(string)(NormalizeDouble((winning/total)*100, 2)) +"\n";
    msg+="Starting from "+TimeToString(from);
    bot.SendMessage(bot.user_chat, msg);
    return true;
}

int max_divisor(double &lot, int no_of_tps, string symbol)
{

    if(!MM) return no_of_tps;
    double min_lot=0.000001 +SymbolInfoDouble(symbol, SYMBOL_VOLUME_MIN);
    double max_lot=SymbolInfoDouble(symbol, SYMBOL_VOLUME_MAX);
    double vol=(lot/min_lot);
    int n_t=(int)(vol);
    if(n_t>no_of_tps) lot =(vol/no_of_tps)*min_lot;
    else if(n_t>1) lot =(vol/n_t)*min_lot;
    else if(lot>max_lot) lot=max_lot;
    string text="Lot Size: " +(string)(lot)+" Number of tps: "+(string)(no_of_tps)+" Number of positions: "+(string)(n_t);
    lot=NormalizeDouble(lot,CountDecimalPlaces(SymbolInfoDouble( symbol, SYMBOL_VOLUME_STEP)));
    bot.SendMessage(bot.user_chat,text);
    if(n_t>no_of_tps) return no_of_tps;
    return n_t;
}

double GetDoubleValue(string text)
{
    string pieces[];
    double price=0;
    StringSplitTrim(text," ",pieces);
    int n=ArraySize(pieces);
    while(--n>=0) {if(StringIsNumeric(pieces[n])) {price=(double)(pieces[n]); break;}}
    return price;
}

void init_filters()
{
    string pair[];
    StringSplitTrim(tradables,",", pair);
    for(int i=0; i<ArraySize(pair); i++)
    {
        string symbol=Instrument(FindW(pair[i]), pair[i]);
        if(symbol!="") tradable_pairs+=" , " + symbol;
    }
    StringSplitTrim(untradables,",", pair);
    for(int i=0; i<ArraySize(pair); i++)
    {
        string symbol=Instrument(FindW(pair[i]), pair[i]);
        if(symbol!="") non_tradable_pairs+=" , " + symbol;
    }
    //Symbol mapping.
    StringSplitTrim(symbol_map,",", pair);
    for(int i=0; i<ArraySize(pair); i++)
    {
        string symbol=Instrument(FindW(pair[i]), pair[i]);
        if(symbol!="")
        {
            int key_index=ArraySize(pair_key);
            ArrayResize(pair_key, key_index+1);
            int attribute_index=ArraySize(pair_attribute);
            ArrayResize(pair_attribute, attribute_index+1);
            pair_key[key_index]=symbol;
            pair_attribute[attribute_index]=pair[i];
        }
    }
    //
}